import functools

from jax import numpy as jnp
import jax
from jax import lax

@functools.partial(jax.jit, static_argnums=(5, 6))
def bart_correlation_maxd_recursive(nminus, n0, nplus, alpha, beta, maxd, d):
    
    p = len(nminus)

    if d >= maxd:
        return 1.
        
    elif d + 1 >= maxd:
        nout = nminus + nplus
        sump = jnp.sum(jnp.where(n0, nout / (nout + n0), 1))
    
    # TODO hardcode d + 2 case
    
    else:
        val = (0., nminus, n0, nplus)
        def loop(i, val):
            sump, nminus, n0, nplus = val
    
            nminusi = nminus[i]
            n0i = n0[i]
            nplusi = nplus[i]
            
            val = (0., nminus, n0, nplus, i, nminusi)
            def loop(k, val):
                sumn, nminus, n0, nplus, i, nminusi = val
                
                nminus = nminus.at[jnp.where(k < nminusi, i, i + p)].set(k)
                nplus = nplus.at[jnp.where(k >= nminusi, i, i + p)].set(k - nminusi)
                
                sumn += bart_correlation_maxd_recursive(nminus, n0, nplus, alpha, beta, maxd, d + 1)
                
                nminus = nminus.at[i].set(nminusi)
                nplus = nplus.at[i].set(nplusi)
                
                return sumn, nminus, n0, nplus, i, nminusi
            
            sumn, nminus, n0, nplus, _, _ = lax.fori_loop(0, nminusi + nplusi, loop, val)
    
            sump += jnp.where(n0i, sumn / (nminusi + n0i + nplusi), 1)
    
            return sump, nminus, n0, nplus

        sump, _, _, _ = lax.fori_loop(0, p, loop, val)

    pnt = alpha / (1 + d) ** beta
    return 1 - pnt * (1 - sump / p)

bart_correlation_maxd_vectorized = jax.vmap(bart_correlation_maxd_recursive, [
    0, 0, 0, None, None, None, None,
])
    
def bart_correlation_maxd(splitsbefore, splitsbetween, splitsafter, alpha, beta, maxd):
    """
    Compute the BART prior correlation between two points.
    
    The correlation is computed approximately by limiting the maximum depth
    of the trees. Limiting trees to depth 1 is equivalent to setting beta to
    infinity.
        
    Parameters
    ----------
    splitsbefore : int (..., p) array
        The number of splitting points less than the two points, separately
        along each coordinate.
    splitsbetween : int (..., p) array
        The number of splitting points between the two points, separately along
        each coordinate.
    splitsafter : int (..., p) array
        The number of splitting points greater than the two points, separately
        along each coordinate.
    alpha, beta : scalar
        The hyperparameters of the branching probability.
    maxd : int
        The maximum depth of the trees. The root has depth zero.
    
    Return
    ------
    corr : scalar
        The prior correlation.
    """
    args = [splitsbefore, splitsbetween, splitsafter]
    args = list(map(jnp.asarray, args))
    for a in args:
        assert a.ndim >= 1
    p = args[0].shape[-1]
    for a in args:
        assert a.shape[-1] == p
    args = jnp.broadcast_arrays(*args)
    shape = args[0].shape[:-1]
    args = [a.reshape(-1, p) for a in args]
    out = bart_correlation_maxd_vectorized(*args, alpha, beta, maxd, 0)
    return out.reshape(shape)



from matplotlib import pyplot as plt
import numpy as np

import lsqfitgp as lgp

fig, axs = plt.subplots(1, 2, clear=True, num='matern')

x = np.linspace(0, 10, 1000)
y0 = lgp.ExpQuad()(0, x)
axs[1].plot(x, y0, label='ExpQuad')

for v in range(1, 100, 10):
    x = np.logspace(-15, 15, 10000)
    y0 = lgp.ExpQuad()(0, x)
    y1 = lgp.Matern(nu=v)(0, x)
    diff = np.abs(y0 - y1)
    axs[0].plot(x, diff, label=f'$\\nu = {v}$')
    
    x = np.linspace(0, 10, 1000)
    y1 = lgp.Matern(nu=v)(0, x)
    axs[1].plot(x, y1, label=f'$\\nu = {v}$')    

ax = axs[0]
ax.set_xscale('log')
ax.set_yscale('log')

for ax in axs:
    ax.legend()
    ax.grid()

fig.show()



# lsqfitgp/tests/test_fit.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import sys

import numpy as np
from jax import numpy as jnp
import gvar
from scipy import stats
import pytest

sys.path.insert(0, '.')
import lsqfitgp as lgp

import util

def flat(g):
    """convert dictionary or array to 1D array"""
    if hasattr(g, 'buf'):
        return g.buf
    elif hasattr(g, 'keys'):
        return gvar.BufferDict(g).buf
    else:
        return np.reshape(g, -1)

def quad(A, v):
    """compute v.T @ A^-1 @ v"""
    w, u = np.linalg.eigh(A)
    utv = u.T @ v
    eps = len(A) * 1e-12 * np.max(w)
    return (utv.T / np.maximum(w, eps)) @ utv
    # TODO maybe low-rank is better?

def chisq_test(g, alpha):
    """chisquare test on g being 0"""
    g = flat(g)
    mean = gvar.mean(g)
    cov = gvar.evalcov(g)
    q = quad(cov, mean)
    n = len(mean)
    assert stats.chi2(n).sf(q) > alpha / 2
    assert stats.chi2(n).cdf(q) > alpha / 2

def check_fit(hyperprior, gpfactory, alpha=1e-5):
    """do a fit with empbayes_fit and check the fitted hyperparameters
    are compatible with the ones used to generate the data"""
    
    # generate hyperparameters
    truehp = gvar.sample(hyperprior)
    
    # generate data
    gp = gpfactory(truehp)
    data = gvar.sample(gp.prior())
        
    # run fit
    fit = lgp.empbayes_fit(hyperprior, gpfactory, data, raises=False)
    
    # check fit result against hyperparameters
    chisq_test(fit.p - truehp, alpha)

@util.tryagain
def test_period():
    hp = {
        'log(scale)': gvar.log(gvar.gvar(1, 0.1))
    }
    x = np.linspace(0, 6, 10)
    def gpfactory(hp):
        gp = lgp.GP(lgp.Periodic(scale=hp['scale']))
        gp.addx(x, 'x')
        return gp
    for _ in range(10):
        check_fit(hp, gpfactory)

@util.tryagain
def test_scale():
    hp = {
        'log(scale)': gvar.log(gvar.gvar(3, 0.2))
    }
    x = np.linspace(0, 2 * np.pi * 5, 20)
    def gpfactory(hp):
        gp = lgp.GP(lgp.ExpQuad(scale=hp['scale']))
        gp.addx(x, 'x')
        return gp
    for _ in range(10):
        check_fit(hp, gpfactory, alpha=1e-7)

@util.tryagain
def test_sdev():
    hp = {
        'log(sdev)': gvar.log(gvar.gvar(1, 1))
    }
    x = np.linspace(0, 5, 10)
    def gpfactory(hp):
        gp = lgp.GP(lgp.ExpQuad() * hp['sdev'] ** 2)
        gp.addx(x, 'x')
        return gp
    for _ in range(10):
        check_fit(hp, gpfactory, alpha=1e-8)
    # TODO once I've seen the chi2 check fail with sf(q) = 1e-8. Is this
    # a minimization problem or the posterior distribution of log(sdev) which
    # has a very heavy tail? The distribution of sdev**2 should be a scaled
    # inverse chisquared, so very heavy tailed indeed

def test_flat():
    hp = gvar.BufferDict({
        'log(sdev)': gvar.log(gvar.gvar(1, 1))
    })
    x = np.linspace(0, 5, 10)
    def gpfactory1(hp):
        gp = lgp.GP(lgp.ExpQuad() * hp['sdev'] ** 2)
        gp.addx(x, 'x')
        return gp
    def gpfactory2(hp):
        gp = lgp.GP(lgp.ExpQuad() * jnp.exp(hp[0]) ** 2)
        gp.addx(x, 'x')
        return gp
    def gpfactory3(hp):
        gp = lgp.GP(lgp.ExpQuad() * jnp.exp(hp) ** 2)
        gp.addx(x, 'x')
        return gp
    truehp = gvar.sample(hp)
    truegp = gpfactory1(truehp)
    trueprior = truegp.prior()
    data = gvar.sample(trueprior)
    kw = dict(raises=False)
    fit1 = lgp.empbayes_fit(hp, gpfactory1, data, **kw)
    fit2 = lgp.empbayes_fit(hp.buf, gpfactory2, data, **kw)
    fit3 = lgp.empbayes_fit(hp.buf[0], gpfactory3, data, **kw)
    util.assert_similar_gvars(fit1.p.buf[0], fit2.p[0], fit3.p)

@util.tryagain
def test_method():
    
    hp = gvar.BufferDict({
        'log(sdev)': gvar.log(gvar.gvar(1, 1))
    })
    x = np.linspace(0, 5, 10)
    def gpfactory(hp):
        gp = lgp.GP(lgp.ExpQuad() * hp['sdev'] ** 2)
        gp.addx(x, 'x')
        return gp
    truehp = gvar.sample(hp)
    truegp = gpfactory(truehp)
    trueprior = truegp.prior()
    data_fixed = gvar.sample(trueprior)
    def data_variable(hp):
        return {k: v + hp['log(sdev)'] for k, v in data_fixed.items()}
    
    for data in [data_fixed, data_variable]:
        fits = []
        kws = [
            dict(method='nograd', minkw=dict(options=dict(xatol=1e-6))),
            dict(method='gradient'),
            dict(method='fisher'),
        ]
        if not sys.version.startswith('3.8'):
            kws.append(dict(method='gradient', minkw=dict(method='l-bfgs-b')))

        for kw in kws:
            kwargs = dict(data=data)
            kwargs.update(kw)
            kwargs.setdefault('minkw', {}).update(x0=truehp.buf)
            fit = lgp.empbayes_fit(hp, gpfactory, **kwargs)
            fits.append(fit)
        p = fits[0].minresult.x
        for fit in fits[1:]:
            util.assert_allclose(fit.minresult.x, p, atol=1e-5)

def test_checks():
    with pytest.raises(KeyError):
        lgp.empbayes_fit(gvar.gvar(0, 1), lambda: None, lambda: None, method='cippa')
    with pytest.raises(RuntimeError) as err:
        def makegp(x):
            gp = lgp.GP(lgp.ExpQuad(), checkfinite=False, checksym=False, checkpos=False)
            gp.addx(x * np.nan, 0)
            return gp
        lgp.empbayes_fit(gvar.gvar(0, 1), makegp, {0: 0})
    assert 'minimization failed: ' in str(err.value)

def test_data():
    
    hp = gvar.BufferDict({
        'log(sdev)': gvar.log(gvar.gvar(1, 1))
    })
    x = np.linspace(0, 5, 10)
    def gpfactory(hp):
        gp = lgp.GP(lgp.ExpQuad() * hp['sdev'] ** 2)
        gp.addx(x, 'x')
        return gp
    truehp = gvar.sample(hp)
    truegp = gpfactory(truehp)
    trueprior = truegp.prior()
    
    def makeerr(bd, err):
        return gvar.BufferDict(bd, buf=np.full_like(bd.buf, err))
    
    data_noerr = gvar.sample(trueprior)
    error = makeerr(data_noerr, 0.1)
    zeroerror = makeerr(data_noerr, 0)
    zerocov = gvar.evalcov(gvar.gvar(data_noerr, zeroerror))
    data_err = gvar.make_fake_data(gvar.gvar(data_noerr, error))
    
    datas = [
        [
            data_noerr,
            gvar.gvar(data_noerr),
            (data_noerr,),
            (data_noerr, zerocov),
            lambda _: data_noerr,
            lambda _: gvar.gvar(data_noerr),
            lambda _: (data_noerr,),
            lambda _: (data_noerr, zerocov),
        ],
        [
            data_err,
            (data_err,),
            (gvar.mean(data_err), gvar.evalcov(data_err)),
            lambda _: data_err,
            lambda _: (data_err,),
            lambda _: (gvar.mean(data_err), gvar.evalcov(data_err)),
        ],
    ]
    
    for datasets in datas:
        fits = []
        for data in datasets:
            fit = lgp.empbayes_fit(hp, gpfactory, data)
            fits.append(fit)

        p = fits[0].minresult.x
        for fit in fits[1:]:
            util.assert_allclose(fit.minresult.x, p, atol=1e-6)
            
#### XFAILS #####

# TODO Investigate why this fails, redo the fit in a standalone script.
# It is not wrong harmonic. Is the Laplace approximation not appropriate for
# this model? Or is the hessian approximation with the BFGS matrix wrong?
pytest.mark.xfail(test_period)



import numpy as np
from matplotlib import pyplot as plt
import mpmath
from scipy import special

import lsqfitgp as lgp

zeta = np.vectorize(lambda *args: float(mpmath.zeta(*args)))

fig, ax = plt.subplots(num='zeta3', clear=True)

s = -1 + 1e-4 * np.linspace(-1, 1, 1001)
z = special.zeta(s)
zsure = zeta(s)
ax.plot(s, np.abs(z - zsure) / np.max(np.abs(zsure)))

ax.set_yscale('log')
ax.set_xlabel('s')
ax.set_title('$|\\zeta_\\mathrm{scipy}(s) - \\zeta(s)|\\,/\\,|\\zeta(s)|$')
ax.axhline(100 * np.finfo(float).eps, linestyle='--', color='red', label='100 ULP')
ax.axhline(np.finfo(float).eps, linestyle='--', color='black', label='1 ULP')
ax.legend()

fig.show()



# lsqfitgp/examples/x.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

                            EXAMPLE X.

    Where the derivatives of an interesting correlation function
    are put to harsh a trial.

"""

import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np

fig, ax = plt.subplots(num='x', clear=True)

x = np.linspace(0, 10, 1000)

eps = 1e-8
args = [
    # Q, kw
    (eps, {}),
    (0.5 - eps, {}),
    (0.5 + eps, dict(linestyle='--')),
    (1 - eps, {}),
    (1 + eps, dict(linestyle='--')),
    (2, {})
]
for Q, kw in args:
    y = lgp.Harmonic(Q=Q).diff(1, 1)(0, x)
    ax.plot(x, y, label=f'Q={Q}', **kw)

ax.legend(loc='best')
fig.show()



# lsqfitgp/_linalg/_decomp.py
#
# Copyright (c) 2020, 2022, 2023 Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

Decompositions of nonnegative definite matrices.

A decomposition object is initialized with a square symmetric matrix and then
can solve linear systems and do other stuff with that matrix. The matrix is
assumed to be positive semidefinite in principle but not numerically,
decompositions differ on how the eventual numerical degeneracy is handled.

It is intended that the matrix inverse is a Moore-Penrose pseudoinverse in case
of (numerically) singular matrices. Some decompositions do not support
pseudoinversion, i.e., the matrix must be well-conditioned.

Abstract classes
----------------
Decomposition
    Abstract base class.
DecompPyTree
    Abstract subclass that register subclasses as jax pytrees.
DecompAutoDiffBase
    Abstract subclass that adds decomposition-independent jax-autodiff support.
DecompAutoDiff
    Specialization of DecompAutoDiffBase for classes which take as
    initialization input a single matrix object.
Diag
    Diagonalization.
LowRankDiag
    Diagonalization with spectrum truncation.

Concrete classes
----------------
EigCutFullRank
    Diagonalization rounding up small (or negative) eigenvalues.
EigCutLowRank
    Diagonalization removing small (or negative) eigenvalues.
SVDCutFullRank
    Diagonalization rounding up small eigenvalues, keeping their sign.
SVDCutLowRank
    Diagonalization removing small eigenvalues.
Lanczos, LOBPCG
    Partial diagonalization with higher eigenvalues only.
Chol
    Cholesky regularized using an estimate of the maximum eigenvalue.
CholToeplitz
    Cholesky decomposition of a Toeplitz matrix regularized using an estimate
    of the maximum eigenvalue. Uses Schur's and Levisons' algorithms. Does not
    store the Cholesky factor in memory.

Composite decompositions
------------------------
Block
    Decompose a 2x2 partitioned matrix.
BlockDiag
    Decompose a 2x2 block diagonal matrix.
SandwichQR, SandwichSVD
    Decompose BAB'.
Woodbury, Woodbury2
    Decompose A ± BCB'.
Pinv, Pinv2
    Turn any decomposition into one that supports pseudoinversion.

"""

import abc
import functools
import warnings
import inspect

import jax
from jax import numpy as jnp
from jax.scipy import linalg as jlinalg
import numpy
from scipy import linalg
from scipy import sparse
import gvar

from . import _toeplitz
from .. import _patch_jax
from . import _pytree

# TODO optimize matrix multiplication with gvars. Use these gvar internals:
# gvar.svec(int size)
# gvar.svec._assign(float[] values, int[] indices)
# gvar.GVar(float mean, svec derivs, smat cov)
# it may require cython to be fast since it's not vectorized

def _transpose(x):
    """ swap the last two axes of array x, corresponds to matrix tranposition
    with the broadcasting convention of matmul """
    if x.ndim < 2:
        return x
    elif isinstance(x, jnp.ndarray):
        return jnp.swapaxes(x, -2, -1)
    else:
        # need to support numpy because this function is used with gvars
        return numpy.swapaxes(x, -2, -1)

class Decomposition(metaclass=abc.ABCMeta):
    """
    
    Abstract base class for nonnegative definite symmetric matrix decomposition.
    
    Methods
    -------
    solve
    quad
    diagquad
    logdet
    tracesolve
    correlate
    decorrelate
    inv
    
    Properties
    ----------
    n
    m
    
    """
    
    @abc.abstractmethod
    def __init__(self, *args, **kw): # pragma: no cover
        """
        Decompose a symmetric nonnegative definite matrix K.
        """
        raise NotImplementedError
        
    @abc.abstractmethod
    def solve(self, b): # pragma: no cover
        """
        Compute K⁺b.
        """
        raise NotImplementedError
    
    def quad(self, b, c=None):
        """
        Compute the quadratic form b'K⁺b if c is not specified, else b'K⁺c.
        `c` can be an array of gvars.
        """
        if c is None:
            return _transpose(b) @ self.solve(b)
        elif c.dtype == object:
            return numpy.asarray(_transpose(self.solve(b))) @ c
        else:
            return _transpose(b) @ self.solve(c)
    
    def diagquad(self, b):
        """
        Compute diag(b'K⁺b).
        """
        result = self.quad(b)
        if result.ndim >= 2:
            result = jnp.diagonal(result, 0, -2, -1)
        return result
        
    # TODO to compute efficiently the likelihood with things like toeplitz and
    # other memoryless solvers to be added in the future, add quad_and_logdet.
    
    @abc.abstractmethod
    def logdet(self): # pragma: no cover
        """
        Compute log det K.
        """
        raise NotImplementedError
        
        # TODO adopt a clear convention: if the matrix is degenerate, it's
        # log pdet + regularization to make it continuous => problem: when the
        # regularization changes with the maximum eigenvalue, this gives a
        # large logdet variation which does not make sense. The regularization
        # should be fixed. Whatevs, the derivatives won't take this variation
        # into account. Maybe I should use a fixed eps? Could that be one of
        # the sources of problems in the hyperparameter fit, e.g., the periodic
        # fit in the tests? Alternative: use a separate fixed eps for the
        # regularization. This does not give continuity, but should give
        # coherent comparability both at the same and different ranks, if the
        # fixed eps is small enough. Alternative: is there a way to correct
        # dynamically the regularization based on the previous ones? The decomp
        # could provide some opaque reg info object, that is to be optionally
        # fed to logdet. Maybe it can be made decomposition-agnostic.
        #
        # => change the current methods' name to logpdet, then make a new one
    
    def tracesolve(self, b):
        """
        Compute tr(K⁺b).
        """
        # return jnp.trace(self.solve(b))
        return jnp.einsum('ij,ji', self.inv(), b)
    
    @abc.abstractmethod
    def correlate(self, b, *, transpose=False): # pragma: no cover
        """
        Compute Ab where K = AA', with A n x m. If b represents iid
        variables with unitary variance, Ab has covariance matrix K. If
        transpose=True, compute A'b.
        """
        raise NotImplementedError
    
    def decorrelate(self, b, *, transpose=False):
        """
        Compute A⁺b, where K = AA', with A n x m. If b represents variables
        with covariance matrix K, A⁺b has idempotent covariance A⁺A. If
        `transpose=True`, compute A⁺'b.
        """
        # K = AA'
        # A⁺ = A'(AA')⁺ = A'K⁺
        # A⁺' = K⁺A
        if transpose:
            return self.solve(self.correlate(b))
        else:
            return self.correlate(self.solve(b), transpose=True)
    
    def inv(self):
        """
        Invert the matrix.
        """
        
        # TODO currently no subclass overrides this, some should probably do.
        
        return self.quad(jnp.eye(self.n))
    
    @property
    def n(self):
        """
        The size of the matrix.
        """
        return len(self.matrix())
    
    @property
    def m(self):
        """
        The inner size of the decomposition used by `correlate` and
        `decorrelate`.
        """
        return self.n
    
    @abc.abstractmethod
    def matrix(self): # pragma: no cover
        """
        Return K.
        """
        raise NotImplementedError
    
    # TODO new properties: rank is the rank for low-rank or rank revealing
    # decompositions, otherwise n.
    
    def _parseeps(self, mat, epsrel, epsabs, maxeigv=None):
        machine_eps = jnp.finfo(_patch_jax.float_type(mat)).eps
        if epsrel == 'auto':
            epsrel = len(mat) * machine_eps
        if epsabs == 'auto':
            epsabs = machine_eps
        if maxeigv is None:
            maxeigv = _gershgorin_eigval_bound(mat)
        self._eps = epsrel * maxeigv + epsabs
        return self._eps
    
    @property
    def eps(self):
        """
        The threshold below which eigenvalues are too small to be determined.
        """
        return self._eps

class DecompPyTree(Decomposition, _pytree.AutoPyTree):
    """
    Marks a decomposition as JAX PyTree, with automatic detection of which
    members are to be traced by JAX.
    """
    pass

class DecompAutoDiffBase(DecompPyTree):
    """
    Abstract subclass adding JAX autodiff support to subclasses of
    Decomposition, even if the decomposition algorithm is not supported by JAX.
    Concrete subclasses have to define _matrix and _map_traceable_init_args.
    """
    
    @abc.abstractmethod
    def _matrix(self, *args, **kw): # pragma: no cover
        """ Given the initialization arguments, return the matrix to be
        decomposed """
        raise NotImplementedError
            
    @abc.abstractmethod
    def _map_traceable_init_args(self, fun, *args, **kw): # pragma: no cover
        """
        
        Given a function Any -> Any, and the initialization arguments,
        apply the function only to those arguments which are considered inputs
        to the definition of the matrix. The return layout must be stable
        under repeated application, i.e., after
        
            a1, k1 = self._map_traceable_init_args(lambda x: x, *a0, **k0)
            a2, k2 = self._map_traceable_init_args(lambda x: x, *a1, **k1)
        
        it must be a1 == a2 and k1 == k2.
        
        """
        raise NotImplementedError
        
    def _store_args(self, args, kw):
        """ store initialization arguments """
        class Marker: pass
        args, kw = self._map_traceable_init_args(lambda x: x, *args, **kw)
        # apply _map_traceable_init_args twice to have a consistent placement
        # of arguments between args, kw and margs, mkw
        margs, mkw = self._map_traceable_init_args(lambda _: Marker, *args, **kw)
        self._dad_jax_args_ = {
            i: v for i, (v, m) in enumerate(zip(args, margs)) if m is Marker
        }
        self._dad_other_args_ = {
            i: v for i, (v, m) in enumerate(zip(args, margs)) if m is not Marker
        }
        self._dad_jax_kw_ = {
            k: v for k, v in kw.items() if mkw[k] is Marker
        }
        self._dad_other_kw_ = {
            k: v for k, v in kw.items() if mkw[k] is not Marker
        }
    
    def _jax_vars(self):
        """ tell AutoPyTree about additional variables to be considered
        children """
        return super()._jax_vars() + ['_dad_jax_args_', '_dad_jax_kw_']
    
    def _get_args(self):
        """ get initialization arguments saved by _store_args """
        args = dict(self._dad_jax_args_)
        args.update(self._dad_other_args_)
        args = tuple(args[i] for i in sorted(args))
        kw = dict(**self._dad_jax_kw_, **self._dad_other_kw_)
        return args, kw
    
    def matrix(self):
        args, kw = self._get_args()
        return self._matrix(*args, **kw)
        
    def _getarg(self, kw, attr, default, nextguy):
        """ saves an initialization argument as instance attribute, and decide
        if the argument is to be passed to the subclass __init__ """
        if not hasattr(self, attr):
            setattr(self, attr, kw.get(attr, default))
            if not self._hasarg(nextguy, attr):
                kw.pop(attr, None)
    
    @staticmethod
    def _hasarg(func, arg):
        """ check if callable func accepts argument keyword argument arg """
        try:
            inspect.signature(func).bind_partial(**{arg: None})
        except TypeError:
            return False
        else:
            return True
    
    # TODO try using jax.linearize and jax.convert_closure. Or: try to pass
    # the decomposition products as variadic arguments to the intermediate
    # wrapper, maybe this solves all reverse-mode autodiff problems like
    # the double undefined primals.
        
    def __init_subclass__(cls, **kw):
        """ wrap the subclass methods to implement automatic differentiation """
        
        super().__init_subclass__(**kw)
        
        old__init__ = cls.__init__

        if inspect.isabstract(cls):
            # Do not wrap the class methods if it's still abstract, because
            # __init_subclass__ will necessarily be invoked again on the
            # concrete subclasses. But wrap the __init__ to drop
            # DecompAutoDiffBase additional arguments eventually passed to
            # super().__init__ from the subclass.
            
            argnames = ['direct_autodiff', 'stop_hessian']
            missing_names = [
                name for name in argnames
                if not cls._hasarg(old__init__, name)
            ]
            if missing_names:
                @functools.wraps(old__init__)
                def __init__(self, *args, **kw):
                    for name in missing_names:
                        kw.pop(name, None)
                    old__init__(self, *args, **kw)
                cls.__init__ = __init__
            
            return
        
        @functools.wraps(old__init__)
        def __init__(self, *args, **kw):
            
            # I can't avoid double wrapping because the superclass __init__ is
            # wrapped and may get called with super(), so I use a flag to check
            # if we are an __init__ called by the old__init__ of a subclass
            if getattr(self, '_dad_initialized_', False):
                return old__init__(self, *args, **kw)
            self._dad_initialized_ = True
            
            self._getarg(kw, 'direct_autodiff', False, old__init__)
            self._getarg(kw, 'stop_hessian', False, old__init__)
            
            if self.direct_autodiff and self.stop_hessian:
                nargs, nkw = self._map_traceable_init_args(_patch_jax.stop_hessian, *args, **kw)
                old__init__(self, *nargs, **nkw)
            elif self.direct_autodiff:
                old__init__(self, *args, **kw)
            else:
                nargs, nkw = self._map_traceable_init_args(jax.lax.stop_gradient, *args, **kw)
                old__init__(self, *nargs, **nkw)
                if self.stop_hessian:
                    args, kw = self._map_traceable_init_args(_patch_jax.stop_hessian, *args, **kw)
            
            self._store_args(args, kw)
        
        cls.__init__ = __init__
        
        for name in 'solve', 'quad', 'logdet', 'tracesolve':
            meth = getattr(cls, name)
            if not hasattr(meth, '_autodiff'):
                newmeth = getattr(cls, '_make_' + name)(meth)
                newmeth = functools.wraps(meth)(newmeth)
                if not hasattr(newmeth, '_autodiff'):
                    newmeth._autodiff = True
                setattr(cls, name, newmeth)
    
    @staticmethod
    def _make_solve(oldsolve):
        
        @functools.partial(jax.custom_jvp, nondiff_argnums=(0,))
        def solve_autodiff(self, K, b):
            return oldsolve(self, b)
        
        @solve_autodiff.defjvp
        def solve_vjp(self, primals, tangents):
            assert not self.direct_autodiff
            K, b = primals
            K_dot, b_dot = tangents
            primal = solve_autodiff(self, K, b)
            tangent_K = -solve_autodiff(self, K, K_dot) @ primal
            tangent_b = solve_autodiff(self, K, b_dot)
            return primal, tangent_K + tangent_b
        
        def solve(self, b):
            if self.direct_autodiff:
                return oldsolve(self, b)
            else:
                return solve_autodiff(self, self.matrix(), b)
        
        # solve_autodiff is used by logdet_jvp and quad_jvp.
        solve._autodiff = solve_autodiff
        
        return solve
    
    @staticmethod
    def _make_quad(oldquad):
        
        @functools.partial(jax.custom_jvp, nondiff_argnums=(0,))
        def quad_autodiff(self, K, b, c):
            return oldquad(self, b, c)

        @functools.partial(jax.custom_jvp, nondiff_argnums=(0,))
        def quad_autodiff_cnone(self, K, b):
            return oldquad(self, b)
        
        # TODO the tangent_K calculation with quad_autodiff breaks backward
        # derivatives, I don't know why. Maybe I should define a new
        # decomposition function sandwich(b, X, c) which computes b.T K^-1 X
        # K^-1 c and use it both for numerical accuracy (I'm guessing that
        # avoiding the two concatenated solves within solve is good) and to make
        # the jvp work. For things like ToeplitzML that stream-decompose each
        # time, it would be more efficient to have versions of solve and quad
        # with an arbitrary number of right-hand sides. => I could have
        # a generic operation sandwich(b, X1, X2, ..., c) that computes
        # b K^-1 X1 K^-1 X2 K^-1 ... K^-1 c, such that its derivatives can
        # be defined in terms of itself, but handling repeated arguments is
        # cumbersome.

        @quad_autodiff.defjvp
        def quad_jvp(self, primals, tangents):
            assert not self.direct_autodiff
            K, b, c = primals
            K_dot, b_dot, c_dot = tangents
            primal = quad_autodiff(self, K, b, c)
            # tangent_K = -quad_autodiff(self, K, b, quad_autodiff(self, K, K_dot, c))
            Kb = self.solve._autodiff(self, K, b)
            Kc = self.solve._autodiff(self, K, c)
            tangent_K = -_transpose(Kb) @ K_dot @ Kc
            tangent_b = quad_autodiff(self, K, b_dot, c)
            # tangent_b = _transpose(b_dot) @ Kc
            tangent_c = quad_autodiff(self, K, b, c_dot)
            # tangent_c = _transpose(Kb) @ c_dot
            return primal, tangent_K + tangent_b + tangent_c
            
        @quad_autodiff_cnone.defjvp
        def quad_cnone_jvp(self, primals, tangents):
            assert not self.direct_autodiff
            K, b = primals
            K_dot, b_dot = tangents
            primal = quad_autodiff_cnone(self, K, b)
            # tangent_K = -quad_autodiff(self, K, b, quad_autodiff(self, K, K_dot, b))
            Kb = self.solve._autodiff(self, K, b)
            tangent_K = -_transpose(Kb) @ K_dot @ Kb
            # using solve + matmul instead of double for tangent_K quad does
            # only O(n^2) matrix-vector multiplications instead of O(n^3)
            # matrix-matrix, but my intuition suggests it could be less
            # numerically accurate
            tangent_b = quad_autodiff(self, K, b_dot, b)
            # tangent_b = _transpose(b_dot) @ Kb
            return primal, tangent_K + tangent_b + _transpose(tangent_b)

        def quad(self, b, c=None):
            if self.direct_autodiff:
                return oldquad(self, b, c)
            elif c is None:
                return quad_autodiff_cnone(self, self.matrix(), b)
            elif c.dtype == object:
                return oldquad(self, b, c)
                # if c contains gvars, the output is an array of gvars, and
                # quad_autodiff would hang because custom_jvp functions are
                # checked not to return object dtypes, so I have to call
                # oldquad
            else:
                return quad_autodiff(self, self.matrix(), b, c)

        quad._autodiff = quad_autodiff
        
        return quad
    
    @staticmethod
    def _make_logdet(oldlogdet):
        
        @functools.partial(jax.custom_jvp, nondiff_argnums=(0,))
        def logdet_autodiff(self, K):
            return oldlogdet(self)
                        
        @logdet_autodiff.defjvp
        def logdet_jvp(self, primals, tangents):
            assert not self.direct_autodiff
            K, = primals
            K_dot, = tangents
            primal = logdet_autodiff(self, K)
            # tangent = jnp.trace(self.solve._autodiff(self, K, K_dot))
            tangent = self.tracesolve._autodiff(self, K, K_dot)
            return primal, tangent
        
        def logdet(self):
            if self.direct_autodiff:
                return oldlogdet(self)
            else:
                return logdet_autodiff(self, self.matrix())
        
        return logdet

    @staticmethod
    def _make_tracesolve(oldtracesolve):

        @functools.partial(jax.custom_jvp, nondiff_argnums=(0,))
        def tracesolve_autodiff(self, K, b):
            return oldtracesolve(self, b)
                        
        @tracesolve_autodiff.defjvp
        def tracesolve_jvp(self, primals, tangents):
            assert not self.direct_autodiff
            K, b = primals
            K_dot, b_dot = tangents
            primal = tracesolve_autodiff(self, K, b)
            tangent_K = -tracesolve_autodiff(self, K, self.quad._autodiff(self, K, K_dot, b))
            tangent_b = tracesolve_autodiff(self, K, b_dot)
            return primal, tangent_K + tangent_b
        
        def tracesolve(self, b):
            if self.direct_autodiff:
                return oldtracesolve(self, b)
            else:
                return tracesolve_autodiff(self, self.matrix(), b)

        tracesolve._autodiff = tracesolve_autodiff
        
        return tracesolve

class DecompAutoDiff(DecompAutoDiffBase):
    """ Specialization of DecompAutoDiffBase for classes which take as
    initialization input a single matrix object """
    
    def _matrix(self, K, *_, **__):
        return K
    
    def _map_traceable_init_args(self, fun, K, *args, **kw):
        return (fun(K), *args), kw
        
class Diag(DecompAutoDiff):
    """
    Diagonalization.
    """
    
    @abc.abstractmethod
    def __init__(self, K, *, epsrel='auto', epsabs=0):
        self._w, self._V = jlinalg.eigh(K)
        self._parseeps(K, epsrel, epsabs, jnp.max(jnp.abs(self._w)))
    
    def solve(self, b):
        return (self._V / self._w) @ (self._V.T @ b)
    
    def quad(self, b, c=None):
        VTb = self._V.T @ b
        bTVw = _transpose(VTb) / self._w
        if c is None:
            VTc = VTb
        elif c.dtype == object:
            VTc = numpy.array(self._V).T @ c
            bTVw = numpy.array(bTVw)
        else:
            VTc = self._V.T @ c
        return bTVw @ VTc
    
    def logdet(self):
        return jnp.sum(jnp.log(jnp.abs(self._w)))
    
    def correlate(self, b, *, transpose=False):
        A = self._V * jnp.sqrt(jnp.abs(self._w))
        if transpose:
            A = A.T
        return A @ b
    
    def decorrelate(self, b, *, transpose=False):
        A_1 = (self._V / jnp.sqrt(jnp.abs(self._w))).T
        if transpose:
            A_1 = A_1.T
        return A_1 @ b
        
# TODO merge the diagonalizations into a single class Diag with options
# rank=None, if specified applies before epsilon cut, inflate: bool, signed:
# bool. Default inflate=True (to avoid ignoring incompatible y and breaking the
# continuity of the logdet), signed=True (to provide an accurate solve and
# quad). In the GP interface, make a single method name 'diag'.
    
class EigCutFullRank(Diag):
    """
    Diagonalization. Eigenvalues below `eps` are set to `eps`, where `eps` is
    relative to the largest eigenvalue.
    """
    
    def __init__(self, K, **kw):
        super().__init__(K, **kw)
        self._w = jnp.where(self._w < self.eps, self.eps, self._w)
            
class EigCutLowRank(Diag):
    """
    Diagonalization. Eigenvalues below `eps` are removed, where `eps` is
    relative to the largest eigenvalue.
    """
    
    def __init__(self, K, **kw):
        super().__init__(K, **kw)
        cond = self._w < self.eps
        self._w = jnp.where(cond, 1, self._w)
        self._V = jnp.where(cond, 0, self._V)
        
        # TODO rank parameter to fix regularization

class SVDCutFullRank(Diag):
    """
    Diagonalization. Eigenvalues below `eps` in absolute value are set to
    `eps` with their sign, where `eps` is relative to the largest eigenvalue.
    """
    
    def __init__(self, K, **kw):
        super().__init__(K, **kw)
        cond = jnp.abs(self._w) < self.eps
        self._w = jnp.where(cond, self.eps * jnp.sign(self._w), self._w)
        
class SVDCutLowRank(Diag):
    """
    Diagonalization. Eigenvalues below `eps` in absolute value are removed,
    where `eps` is relative to the largest eigenvalue.
    """
    
    def __init__(self, K, **kw):
        super().__init__(K, **kw)
        cond = jnp.abs(self._w) < self.eps
        self._w = jnp.where(cond, 1, self._w)
        self._V = jnp.where(cond, 0, self._V)

class LowRankDiag(Diag):
    """
    Diagonalization with a fixed size truncation of the spectrum.
    """
    
    @abc.abstractmethod
    def __init__(self, K, *, rank=None):
        if rank is None:
            raise ValueError('rank not specified')
        self._rank = rank
    
    @property
    def m(self):
        return self._rank

class Lanczos(LowRankDiag):
    """
    Keep only the first `rank` higher eigenmodes.
    """
    
    def __init__(self, K, *, rank=None):
        super().__init__(K, rank=rank)
        class wdummy:
            dtype = K.dtype
            shape = (rank,)
        class Vdummy:
            dtype = K.dtype
            shape = (len(K), rank)
        self._w, self._V = jax.pure_callback(
            lambda K: sparse.linalg.eigsh(numpy.asarray(K), k=rank, which='LM'),
            (wdummy, Vdummy), K,
        )
        
        # TODO try using jnp.matmul instead of passing the matrix and check
        # if it improves performance due to lower default comput precision =>
        # behavior under jit not clear
        
        # TODO try using linalg.blas.get_blas_func('symv', [K]) for the matrix
        # vector product. sparse.linalg does not seem to take into account
        # that the matrix is symmetric to do the matvec product.
        
class LOBPCG(LowRankDiag):
    """
    Keep only the first `rank` higher eigenmodes.
    """
    
    def __init__(self, K, *, rank=None):
        super().__init__(K, rank=rank)
        class wdummy:
            dtype = K.dtype
            shape = (rank,)
        class Vdummy:
            dtype = K.dtype
            shape = (len(K), rank)
        self._w, self._V = jax.pure_callback(
            lambda K: sparse.linalg.lobpcg(numpy.asarray(K), numpy.random.randn(*Vdummy.shape)),
            (wdummy, Vdummy), K,
        )
        
        # TODO try using jax's own implementation of LOBPCG

# TODO look into the pivoted Cholesky decomposition from Gardner et al. 2018

def solve_triangular(a, b, lower=False):
    """
    Pure python implementation of scipy.linalg.solve_triangular for when
    a or b are object arrays.
    """
    # TODO maybe commit this to gvar.linalg
    a = numpy.asarray(a)
    x = numpy.copy(b)
    
    vec = x.ndim < 2
    if vec:
        x = x[:, None]

    n = a.shape[-1]
    assert x.shape[-2] == n
    
    if not lower:
        a = a[..., ::-1, ::-1]
        x = x[..., ::-1, :]

    x[..., 0, :] /= a[..., 0, 0, None]
    for i in range(1, n):
        x[..., i:, :] -= x[..., None, i - 1, :] * a[..., i:, i - 1, None]
        x[..., i, :] /= a[..., i, i, None]
    
    if not lower:
        x = x[..., ::-1, :]
    
    if vec:
        x = numpy.squeeze(x, -1)
    return x

def _gershgorin_eigval_bound(mat):
    """
    Upper bound on the largest magnitude eigenvalue of the matrix.
    """
    return jnp.max(jnp.sum(jnp.abs(mat), axis=1))

# TODO consider scipy.linalg.lapack.dpstrf that does cholesky pivoted up to a
# remainder nondef block. on 1000x1000 it takes 2.1x cholesky worst case, i.e.,
# when it decomposes the full matrix, so it's fast. 

class Chol(DecompAutoDiff):
    """
    Cholesky decomposition.
    """
    
    @staticmethod
    def _scale(K):
        """
        Compute a vector s of powers of 2 such that diag(K / outer(s, s)) ~ 1.
        """
        d = jnp.diag(K)
        return jnp.where(d, jnp.exp2(jnp.rint(0.5 * jnp.log2(d))), 1)

    def __init__(self, K, *, epsrel='auto', epsabs=0):
        s = self._scale(K)
        K = K / s / s[:, None]
        eps = self._parseeps(K, epsrel, epsabs)
        K = K.at[jnp.diag_indices(len(K))].add(eps)
        L = jlinalg.cholesky(K, lower=True)
        with _patch_jax.skipifabstract():
            if not jnp.all(jnp.isfinite(L)):
                raise numpy.linalg.LinAlgError('cholesky decomposition not finite, probably matrix not pos def numerically')
        self._L = L * s[:, None]
    
    def solve(self, b):
        invLb = jlinalg.solve_triangular(self._L, b, lower=True)
        return jlinalg.solve_triangular(self._L.T, invLb, lower=False)
    
    def quad(self, b, c=None):
        invLb = jlinalg.solve_triangular(self._L, b, lower=True)
        if c is None:
            invLc = invLb
        elif c.dtype == object:
            invLc = solve_triangular(self._L, c, lower=True)
            invLb = numpy.asarray(invLb)
        else:
            invLc = jlinalg.solve_triangular(self._L, c, lower=True)
        return _transpose(invLb) @ invLc
    
    def logdet(self):
        return 2 * jnp.sum(jnp.log(jnp.diag(self._L)))
    
    def correlate(self, b, *, transpose=False):
        return (self._L.T if transpose else self._L) @ b
    
    def decorrelate(self, b, *, transpose=False):
        L = self._L.T if transpose else self._L
        return jlinalg.solve_triangular(L, b, lower=not transpose)

class CholToeplitz(DecompAutoDiff):
    """
    Cholesky decomposition of a Toeplitz matrix. Only the first row of the
    matrix is read. It does not store the decomposition in memory, it is
    evaluated each time column by column during operations.
    """
    
    def __init__(self, K, *, epsrel='auto', epsabs=0):
        t = jnp.asarray(K[0, :])
        m = _toeplitz.eigv_bound(t)
        eps = self._parseeps(t, epsrel, epsabs, m)
        self.t = t.at[0].add(eps)
    
    def solve(self, b):
        return _toeplitz.solve(self.t, b)
    
    def quad(self, b, c=None):
        t = self.t
        if c is None:
            ilb = _toeplitz.chol_solve(t, b)
            ilc = ilb
        elif c.dtype == object:
            ilb = numpy.array(_toeplitz.chol_solve(t, b))
            ilc = _toeplitz.chol_solve_numpy(t, c)
        else:
            ilb, ilc = _toeplitz.chol_solve(t, b, c)
        return _transpose(ilb) @ ilc
    
    def logdet(self):
        return _toeplitz.logdet(self.t)
    
    def correlate(self, b, *, transpose=False):
        if transpose:
            return _toeplitz.chol_transp_matmul(self.t, b)
        else:
            return _toeplitz.chol_matmul(self.t, b)
    
    def decorrelate(self, b, *, transpose=False):
        if transpose:
            return _toeplitz.chol_transp_solve(self.t, b)
        else:
            return _toeplitz.chol_solve(self.t, b)

class Block(DecompAutoDiffBase):
    """
    Decomposition of a 2x2 symmetric block matrix using decompositions of the
    diagonal blocks.
    """
    
    # For A- it works, see schott 2017, theorem 7.12, page 298
    # For A+, it needs to be modified, see schott 2017, theorem 7.13, page 300
    # For a sufficient condition for working with A+ as written, see schott
    # 2017, exercise 7.28, page 311
    
    def _matrix(self, P_decomp, S, Q, S_decomp_class, **kw):
        assert not self.direct_autodiff
        return jnp.block([[P_decomp.matrix(), Q], [Q.T, S]])
            
    def _map_traceable_init_args(self, fun, P_decomp, S, Q, S_decomp_class, **kw):
        return (fun(P_decomp), fun(S), fun(Q), S_decomp_class), kw

    def __init__(self, P_decomp, S, Q, decompcls, **kw):
        """
        The matrix to be decomposed is
        
            K = [P   Q]
                [Q'  S]
        
        Parameters
        ----------
        P_decomp : Decomposition
            An instantiated decomposition of P.
        S, Q : matrices
            The other blocks.
        decompcls : type
            A subclass of Decomposition used to decompose S - Q'P⁻¹Q.
        **kw :
            Additional keyword arguments are passed to `decompcls`.
        """
        self._Q = Q
        self._invP = P_decomp
        self._tildeS = decompcls(S - P_decomp.quad(Q), **kw)
        
        # TODO does it make sense to allow to use woodbury on the schur
        # complement of P? Maybe to be general I should take an instantiated
        # decomposition schurP_decomp instead of a class. Or maybe detect
        # a woodbury class with a woodbury ABC and pass arguments appropriately.
    
    def solve(self, b):
        invP = self._invP
        tildeS = self._tildeS
        Q = self._Q
        f = b[:len(Q)]
        g = b[len(Q):]
        gQTinvPf = g - invP.quad(Q, f)
        y = tildeS.solve(gQTinvPf)
        x = invP.solve(f - tildeS.quad(Q.T, gQTinvPf))
        return jnp.concatenate([x, y])
    
    def quad(self, b, c=None):
        invP = self._invP
        tildeS = self._tildeS
        Q = self._Q
        f = b[:len(Q)]
        g = b[len(Q):]
        QTinvPf = invP.quad(Q, f)
        if c is None:
            fTinvPQtildeSg = tildeS.quad(QTinvPf, g)
            return invP.quad(f) + tildeS.quad(QTinvPf) - fTinvPQtildeSg - _transpose(fTinvPQtildeSg) + tildeS.quad(g)
        else:
            h = c[:len(Q)]
            i = c[len(Q):]
            QTinvPh = invP.quad(Q, h)
            fTinvPQtildeSi = tildeS.quad(QTinvPf, i)
            gTtildeSQTinvPh = tildeS.quad(g, QTinvPh)
            return invP.quad(f, h) + tildeS.quad(QTinvPf, QTinvPh) - fTinvPQtildeSi - gTtildeSQTinvPh + tildeS.quad(g, i)
    
    def logdet(self):
        return self._invP.logdet() + self._tildeS.logdet()
        # TODO Is this valid for the pseudodeterminant? It isn't in general, but
        # it could be for p.s.d. matrices. schott 2017, theorem 7.10, page 297
        # shows that the rank decomposes in this way, so maybe the pdet does
        # too.
    
    @property
    def eps(self):
        return jnp.maximum(self._invP.eps, self._tildeS.eps())
    
    def correlate(self, b, *, transpose=False):
        # Block Cholesky decomposition:
        # K = [P   Q] = LL'
        #     [Q'  S]
        # L = [A   ]        L' = [A'  B']
        #     [B  C]             [    C']
        # AA' = P
        # AB' = Q  ==>  B' = AQ
        # BB' + CC' = S  ==>  CC' = S - Q'P⁻¹Q = S̃
        invP = self._invP
        tildeS = self._tildeS
        Q = self._Q
        split = len(Q) if transpose else invP.m
        f = b[:split]
        g = b[split:]
        if transpose:
            x = invP.correlate(f, transpose=True) + invP.decorrelate(Q @ g)
            y = tildeS.correlate(g, transpose=True)
        else:
            x = invP.correlate(f)
            y = Q.T @ invP.decorrelate(f, transpose=True) + tildeS.correlate(g)
        return jnp.concatenate([x, y], axis=max(0, b.ndim - 2))
    
    def decorrelate(self, b, *, transpose=False):
        # L⁻¹ = [   A⁻¹       ]     L'⁻¹ = [A'⁻¹  -A'⁻¹B'C'⁻¹]
        #       [-C⁻¹BA⁻¹  C⁻¹]            [          C'⁻¹   ]
        invP = self._invP
        tildeS = self._tildeS
        Q = self._Q
        split = invP.m if transpose else len(Q)
        f = b[:split]
        g = b[split:]
        if transpose:
            y = tildeS.decorrelate(g, transpose=True)
            x = invP.decorrelate(f - invP.decorrelate(Q @ y), transpose=True)
        else:
            x = invP.decorrelate(f)
            y = tildeS.decorrelate(g - invP.quad(Q, f))
        return jnp.concatenate([x, y], axis=max(0, b.ndim - 2))
        
    @property
    def n(self):
        return sum(self._Q.shape)
    
    @property
    def m(self):
        return self._invP.m + self._tildeS.m
    
    # to compute the rank, use schott 2017, theorem 7.10, page 297
    
class BlockDiag(DecompAutoDiffBase):
    
    # TODO allow NxN instead of 2x2?
    
    def _matrix(self, A_decomp, B_decomp):
        assert not self.direct_autodiff
        return jlinalg.block_diag(A_decomp.matrix(), B_decomp.matrix())
            
    def _map_traceable_init_args(self, fun, A_decomp, B_decomp):
        return (fun(A_decomp), fun(B_decomp)), {}

    def __init__(self, A_decomp, B_decomp):
        """
        
        Decomposition of a 2x2 block diagonal matrix.
        
        The matrix is
        
            K = [A   ]
                [   B]
        
        Parameters
        ----------
        A_decomp, B_decomp: Decomposition
            Instantiated decompositions of A and B.
        
        """
        self._A = A_decomp
        self._B = B_decomp
    
    def solve(self, b):
        A = self._A
        B = self._B
        An = A.n
        f = b[:An]
        g = b[An:]
        return jnp.concatenate([A.solve(f), B.solve(g)])
    
    def quad(self, b, c=None):
        A = self._A
        B = self._B
        An = A.n
        f = b[:An]
        g = b[An:]
        if c is None:
            return A.quad(f) + B.quad(g)
        else:
            h = c[:An]
            i = c[An:]
            return A.quad(f, h) + B.quad(g, i)
    
    def logdet(self):
        return self._A.logdet() + self._B.logdet()
    
    @property
    def eps(self):
        return jnp.maximum(self._A.eps, self._B.eps())
    
    def correlate(self, b, *, transpose=False):
        A = self._A
        B = self._B
        An = A.n if transpose else A.m
        f = b[:An]
        g = b[An:]
        return jnp.concatenate([
            A.correlate(f, transpose=transpose),
            B.correlate(g, transpose=transpose),
        ], axis=max(0, b.ndim - 2))
    
    def decorrelate(self, b, *, transpose=False):
        A = self._A
        B = self._B
        An = A.m if transpose else A.n
        f = b[:An]
        g = b[An:]
        return jnp.concatenate([
            A.decorrelate(f, transpose=transpose),
            B.decorrelate(g, transpose=transpose),
        ], axis=max(0, b.ndim - 2))

    @property
    def n(self):
        return self._A.n + self._B.n
    
    @property
    def m(self):
        return self._A.m + self._B.m

class SandwichQR(DecompAutoDiffBase):
    
    def _matrix(self, A_decomp, B):
        assert not self.direct_autodiff
        return B @ A_decomp.matrix() @ B.T
            
    def _map_traceable_init_args(self, fun, A_decomp, B):
        return (fun(A_decomp), fun(B)), {}

    def __init__(self, A_decomp, B):
        """
        Decompose K = BAB' with the QR decomposition of B.
        
        Parameters
        ----------
        A_decomp : Decomposition
            Instantiated decomposition of A.
        B : array
            The B matrix. Must be tall or square.
        
        """
        self._A = A_decomp
        assert B.shape[0] >= B.shape[1]
        self._B = B
        self._q, self._r = jnp.linalg.qr(B, mode='reduced')
        # Do the QR of A.correlate(B.T, T).T and drop _A
    
    def solve(self, b):
        A, q, r = self._A, self._q, self._r
        rqb = jlinalg.solve_triangular(r, q.T @ b, lower=False)
        arqb = A.solve(rqb)
        return q @ jlinalg.solve_triangular(r.T, arqb, lower=True)
    
    def quad(self, b, c=None):
        A, q, r = self._A, self._q, self._r
        rqb = jlinalg.solve_triangular(r, q.T @ b, lower=False)
        if c is None:
            return A.quad(rqb)
        if c.dtype != object:
            rqc = jlinalg.solve_triangular(r, q.T @ c, lower=False)
        else:
            rqc = solve_triangular(r, numpy.matmul(q.T, c), lower=False)
        return A.quad(rqb, rqc)
    
    def logdet(self):
        A, r = self._A, self._r
        B_logdet = jnp.sum(jnp.log(jnp.abs(jnp.diag(r))))
        return A.logdet() + 2 * B_logdet
    
    @property
    def eps(self):
        return self._A.eps # TODO take B into account somehow
    
    def correlate(self, b, *, transpose=False):
        A, B = self._A, self._B
        if transpose:
            return A.correlate(B.T @ b, transpose=True)
        else:
            return B @ A.correlate(b)
    
    def decorrelate(self, b, *, transpose=False):
        # BAB' = BL(BL)'
        # BL = QRL
        # (QRL)⁺ = L⁻¹R⁻¹Q' (valid because Q has orthonormal columns)
        # (QRL)⁺' = QR'⁻¹L'⁻¹
        A, q, r = self._A, self._q, self._r
        if transpose:
            lb = A.decorrelate(b, transpose=True)
            return q @ jlinalg.solve_triangular(r.T, lb, lower=True)
        else:
            rqb = jlinalg.solve_triangular(r, q.T @ b, lower=False)
            return A.decorrelate(rqb)
    
    @property
    def n(self):
        return len(self._B)
    
    @property
    def m(self):
        return self._A.m
    
def _rq(a, **kw):
    """
    Decompose a as r q, with r lower-triangular and q orthogonal
    Return: r, q
    """
    qt, rt = jnp.linalg.qr(_transpose(a), **kw)
    return _transpose(rt), _transpose(qt)

# TODO I am not able to define an efficient pinv of B A B^T with short B.
#
# class SandwichRQ(DecompPyTree):
#
#     def __init__(self, A_decomp, B):
#         """
#         Decompose M = B A B^T with the RQ decomposition of B.
#
#         Parameters
#         ----------
#         A_decomp : Decomposition
#             Instantiated decomposition of A.
#         B : array
#             The B matrix. Must be short or square.
#
#         """
#         self._A = A_decomp
#         assert B.shape[0] <= B.shape[1]
#         self._B = B
#         self._r, self._q = _rq(B, mode='reduced')
#
#     def solve(self, b):
#         A, q, r = self._A, self._q, self._r
#         qrb = q.T @ jlinalg.solve_triangular(r, b, lower=True)
#         aqrb = A.solve(qrb)
#         return jlinalg.solve_triangular(r.T, q @ aqrb, lower=False)
#
#     def quad(self, b, c=None):
#         A, q, r = self._A, self._q, self._r
#         qrb = q.T @ jlinalg.solve_triangular(r, b, lower=True)
#         if c is None:
#             return A.quad(qrb)
#         if c.dtype != object:
#             qrc = q.T @ jlinalg.solve_triangular(r, c, lower=True)
#         else:
#             qrc = numpy.matmul(q.T, solve_triangular(r, c, lower=True))
#         return A.quad(qrb, qrc)
#
#     def logdet(self):
#         A, r = self._A, self._r
#         B_logdet = jnp.sum(jnp.log(jnp.abs(jnp.diag(r))))
#         return A.logdet() + 2 * B_logdet
#         # I don't know if this really makes sense
#
#     def correlate(self, b):
#         A, B = self._A, self._B
#         return B @ A.correlate(b[:A.n])
#
#     def decorrelate(self, b):
#         A, q, r = self._A, self._q, self._r
#         qrb = q.T @ jlinalg.solve_triangular(r, b, lower=True)
#         return A.decorrelate(qrb)
#
#     @property
#     def n(self):
#         return len(self._B)

class SandwichSVD(DecompAutoDiffBase):
    
    # TODO to compute the pseudoinverse, compute C = A.correlate(B.T,
    # transpose=True).T to have M = CCt, then compute C+ to have M+ = C+t C+.
    
    # TODO in a new class: use C+ = (CtC)+ Ct, and decompose CtC with a
    # user-provided class. Make A optional, if A=None, then C = B and M = BBt.
    # If C is short, use C+ = Ct (CCt)+.
    
    def _matrix(self, A_decomp, B):
        assert not self.direct_autodiff
        return B @ A_decomp.matrix() @ B.T
            
    def _map_traceable_init_args(self, fun, A_decomp, B):
        return (fun(A_decomp), fun(B)), {}

    def __init__(self, A_decomp, B):
        """
        Decompose K = BAB' with the SVD decomposition of B.
        
        Parameters
        ----------
        A_decomp : Decomposition
            Instantiated decomposition of A.
        B : array
            The B matrix, must be tall or square.
        
        """
        self._A = A_decomp
        self._B = B
        self._u, self._s, self._vh = jnp.linalg.svd(B, full_matrices=False)
        # TODO regularize s
    
    def solve(self, b):
        A, u, s, vh = self._A, self._u, self._s, self._vh
        vsub = jnp.linalg.multi_dot([vh.T / s, u.T, b])
        vsu = (vh.T / s) @ u.T
        return A.quad(vsu, vsub)
    
    def quad(self, b, c=None):
        A, u, s, vh = self._A, self._u, self._s, self._vh
        vsub = jnp.linalg.multi_dot([vh.T / s, u.T, b])
        if c is None:
            return A.quad(vsub)
        if c.dtype != object:
            vsuc = jnp.linalg.multi_dot([vh.T / s, u.T, c])
        else:
            vsuc = numpy.linalg.multi_dot([vh.T / s, u.T, c])
        return A.quad(vsub, vsuc)
    
    def logdet(self):
        A, s = self._A, self._s
        B_logdet = jnp.sum(jnp.log(s))
        return A.logdet() + 2 * B_logdet # makes sense if A invertible, in
        # general after I svd an empty sandwich
    
    @property
    def eps(self):
        return self._A.eps # TODO take B into account somehow
        
    def correlate(self, b, *, transpose=False):
        A, B = self._A, self._B
        if transpose:
            return A.correlate(B.T @ b, transpose=True)
        else:
            return B @ A.correlate(b)
    
    def decorrelate(self, b, *, transpose=False):
        # BAB' = BL(BL)'
        # BL = USV'L
        # (BL)⁺ = L⁻¹VS⁻¹U'
        # (BL)⁺' = US⁻¹V'L'⁻¹
        A, u, s, vh = self._A, self._u, self._s, self._vh
        if transpose:
            lb = A.decorrelate(b, transpose=True)
            return (u / s) @ (vh @ lb)
        else:
            vsub = (vh.T / s) @ (u.T @ b)
            return A.decorrelate(vsub)
    
    @property
    def n(self):
        return len(self._B)
    
    @property
    def m(self):
        return self._A.m
    
class Woodbury(DecompAutoDiffBase):
    
    def _matrix(self, A_decomp, B, C_decomp, decompcls, sign=1, **kw):
        assert not self.direct_autodiff
        return A_decomp.matrix() + sign * B @ C_decomp.matrix() @ B.T
            
    def _map_traceable_init_args(self, fun, A_decomp, B, C_decomp, decompcls, sign=1, **kw):
        return (fun(A_decomp), fun(B), fun(C_decomp), decompcls, sign), kw

    def __init__(self, A_decomp, B, C_decomp, decompcls, sign=1, **kw):
        """

        Decompose K = A ± BCB' using Woodbury's formula. K, A, C > 0.
        
        Parameters
        ----------
        A_decomp : Decomposition
            Instantiated decomposition of A.
        B : array
            The B matrix, can be rectangular.
        C_decomp : Decomposition
            Instantiated decomposition of C.
        decompcls : type
            Decomposition class used to decompose C⁻¹ ± B'A⁻¹B'.
        sign : {-1, 1}
            The sign in the expression above, default 1.
        **kw :
            Keyword arguments are passed to the initialization of `decompcls`.
        
        """
        assert B.shape == (A_decomp.n, C_decomp.n)
        assert sign in (-1, 1)
        self._A = A_decomp
        self._B = B
        self._C = C_decomp
        X = C_decomp.inv() + sign * A_decomp.quad(B)
        self._X = decompcls(X, **kw)
        self._sign = sign
    
    # (A ± BCB')⁻¹
    #   = A⁻¹ ∓ A⁻¹B (C⁻¹ ± B'A⁻¹B)⁻¹ B'A⁻¹
    #                 ~~~~~~~~~~~~
    #                    =: X
    #   = A⁻¹ ∓ A⁻¹B X⁻¹ B'A⁻¹
    
    def solve(self, b):
        A, B, X, s = self._A, self._B, self._X, self._sign
        ab = A.solve(b)
        AB = A.solve(B)
        bab = A.quad(B, b)
        xb = X.quad(AB.T, bab)
        return ab - s * xb
    
    def quad(self, b, c=None):
        A, B, X, s = self._A, self._B, self._X, self._sign
        if c is None:
            bab = A.quad(b)
            BAB = A.quad(B, b)
            bxb = X.quad(BAB)
            return bab - s * bxb
        else:
            bac = A.quad(b, c)
            BAB = A.quad(B, b)
            BAC = A.quad(B, c)
            bxc = X.quad(BAB, BAC)
            return bac - s * bxc
    
    def logdet(self):
        A, C, X = self._A, self._C, self._X
        return A.logdet() + C.logdet() + X.logdet()
    
    @property
    def eps(self):
        A, C, X = self._A, self._C, self._X
        return jnp.max(jnp.array([A.eps, C.eps, X.eps]))
    
    def correlate(self, b, *, transpose=False):
        # A = LL'
        # C = NN'
        # A + BCB' = LL' + BNN'B' = [L BN] [L BN]'
        assert self._sign > 0
        A, B, C = self._A, self._B, self._C
        if transpose:
            return jnp.concatenate([
                A.correlate(b, transpose=True),
                C.correlate(B.T @ b, transpose=True)
            ], axis=max(0, b.ndim - 2))
        else:
            return A.correlate(b[:A.m]) + B @ C.correlate(b[A.m:])
    
    def decorrelate(self, b, *, transpose=False):
        # V = [L BN]
        # V⁺ = V'(VV')⁻¹
        #    = V'K⁻¹ =
        #    = [L'; N'B'] (A⁻¹ - A⁻¹B X⁻¹ B'A⁻¹) =
        #    = [ L⁻¹ (I - B X⁻¹ B'A⁻¹) ]
        #      [ N'(I - B'A⁻¹B X⁻¹) B'A⁻¹ ]
        # V⁺' = [ (I - A⁻¹B X⁻¹ B') L'⁻¹
        #         A⁻¹ (I - X⁻¹ B'A⁻¹) BN ]
        assert self._sign > 0
        A, B, C, X = self._A, self._B, self._C, self._X
        if transpose:
            f, g = b[:A.m], b[A.m:]
            BtA_1 = A.solve(B).T
            Lt_1f = A.decorrelate(f, transpose=True)
            x = Lt_1f - X.quad(BtA_1, B.T @ Lt_1f)
            BNg = B @ C.correlate(g)
            y = A.solve(BNg - X.quad(B.T, A.quad(B, BNg)))
            return x + y
        else:
            BtA_1b = A.quad(B, b)
            BtA_1B = A.quad(B)
            x = A.decorrelate(b - X.quad(B.T, BtA_1b))
            y = C.correlate(BtA_1b - X.quad(BtA_1B, BtA_1b), transpose=True)
            return jnp.concatenate([x, y], axis=max(0, b.ndim - 2))
    
    @property
    def n(self):
        return self._A.n
    
    @property
    def m(self):
        return self._A.m + self._C.m

class Woodbury2(DecompAutoDiffBase):
    
    def _matrix(self, A_decomp, B, C_decomp, decompcls, sign=1, **kw):
        assert not self.direct_autodiff
        return A_decomp.matrix() + sign * B @ C_decomp.matrix() @ B.T
            
    def _map_traceable_init_args(self, fun, A_decomp, B, C_decomp, decompcls, sign=1, **kw):
        return (fun(A_decomp), fun(B), fun(C_decomp), decompcls, sign), kw

    def __init__(self, A_decomp, B, C_decomp, decompcls, sign=1, **kw):
        """

        Decompose K = A ± BCB' using Woodbury's formula. K, A > 0, C ≥ 0.
                
        Parameters
        ----------
        A_decomp : Decomposition
            Instantiated decomposition of A.
        B : array
            The B matrix, can be rectangular.
        C_decomp : Decomposition
            Instantiated decomposition of C.
        decompcls : type
            Decomposition class used to decompose I ± L'B'A⁻¹BL, where
            C = LL'.
        sign : {-1, 1}
            The sign in the expression above, default 1.
        **kw :
            Keyword arguments are passed to the initialization of `decompcls`.
        
        """
        assert B.shape == (A_decomp.n, C_decomp.n)
        assert sign in (-1, 1)
        self._A = A_decomp
        self._V = C_decomp.correlate(B.T, transpose=True).T
        I = jnp.eye(C_decomp.m)
        self._X = decompcls(I + sign * A_decomp.quad(self._V), **kw)
        self._sign = sign
    
    # C = LL'
    # (A ± BCB')⁻¹ =
    #   = (A ± B L L' B')⁻¹ =
    #          ~~~
    #          =:V
    #   = (A ± VV')⁻¹ =
    #   = A⁻¹ ∓ A⁻¹ V (I ± V' A⁻¹ V)⁻¹ V' A⁻¹
    #                  ~~~~~~~~~~~~
    #                     =: X
    #   = A⁻¹ ∓ A⁻¹ V X⁻¹ V' A⁻¹
    
    # TODO: when the sign is negative, can I replace X⁻¹ with X+ to get K+?
    
    def solve(self, b):
        A, X, V, s = self._A, self._X, self._V, self._sign
        A_1b = A.solve(b)
        VtA_1b = A.quad(V, b)
        A_1V = A.solve(V)
        return A_1b - s * X.quad(A_1V.T, VtA_1b)
    
    def quad(self, b, c=None):
        A, X, V, s = self._A, self._X, self._V, self._sign
        if c is None:
            bTA_1b = A.quad(b)
            VtA_1b = A.quad(V, b)
            return bTA_1b - s * X.quad(VtA_1b)
        else:
            bTA_1c = A.quad(b, c)
            VtA_1c = A.quad(V, c)
            VtA_1b = A.quad(V, b)
            return bTA_1c - s * X.quad(VtA_1b, VtA_1c)
    
    def logdet(self):
        A, X = self._A, self._X
        # det(A ± VV') = det(A) det(I ± V' A⁻¹ V)
        return A.logdet() + X.logdet()
    
    @property
    def eps(self):
        A, X = self._A, self._X
        return jnp.maximum(A.eps, X.eps)
    
    def correlate(self, b, *, transpose=False):
        # A = NN'
        # A + BCB' = NN' + VV' = [N V] [N V]'
        assert self._sign > 0
        A, V = self._A, self._V
        if transpose:
            return jnp.concatenate([
                A.correlate(b, transpose=True),
                V.T @ b,
            ], axis=max(0, b.ndim - 2))
        else:
            return A.correlate(b[:A.m]) + V @ b[A.m:]
    
    def decorrelate(self, b, *, transpose=False):
        # A = NN'
        # A + BCB' = [N V] [N V]'
        #            ~~~~~
        #             =: Q
        # Q+ = Q' (QQ')⁻¹ =
        #    = Q' (A + BCB')⁻¹ =
        #    = [N'; V'] (A⁻¹ - A⁻¹ V X⁻¹ V' A⁻¹) =
        #    = [ N' N'⁻¹ N⁻¹ - N' N'⁻¹ N⁻¹ V X⁻¹ V' A⁻¹ ] =
        #      [ V' A⁻¹ - V' A⁻¹ V X⁻¹ V' A⁻¹           ]
        #    = [ N⁻¹ - N⁻¹ V X⁻¹ V' A⁻¹    ] =
        #      [ (I - V' A⁻¹ V X⁻¹) V' A⁻¹ ]
        #    = [ (I - N⁻¹ V X⁻¹ V' N'⁻¹) N⁻¹         ]
        #      [ (I - V' N'⁻¹ N⁻¹ V X⁻¹) V' N'⁻¹ N⁻¹ ]
        # Q+' = [ N'⁻¹ - A⁻¹ V X⁻¹ V' N'⁻¹
        #         A⁻¹ V - A⁻¹ V X⁻¹ V' A⁻¹ V ] =
        #     = [ (I - A⁻¹ V X⁻¹ V') N'⁻¹
        #         (I - A⁻¹ V X⁻¹ V') A⁻¹ V ]
        assert self._sign > 0
        A, V, X = self._A, self._V, self._X
        if transpose:
            f, g = b[:A.m], b[A.m:]
            Nt_1f = A.decorrelate(f, transpose=True)
            A_1V = A.solve(V)
            x = Nt_1f - X.quad(A_1V.T, V.T @ Nt_1f)
            Vg = V @ g
            VtA_1Vg = A.quad(V, Vg)
            A_1Vg = A.solve(Vg)
            y = A_1Vg - X.quad(A_1V.T, VtA_1Vg)
            return x + y
        else:
            N_1V = A.decorrelate(V)
            N_1b = A.decorrelate(b)
            VtA_1b = N_1V.T @ N_1b
            x = N_1b - X.quad(N_1V.T, VtA_1b)
            VtA_1V = N_1V.T @ N_1V
            y = VtA_1b - X.quad(VtA_1V, VtA_1b)
            return jnp.concatenate([x, y], axis=max(0, b.ndim - 2))
        
    @property
    def n(self):
        return self._A.n
    
    @property
    def m(self):
        return self._A.m + self._V.shape[1]

class Pinv(DecompAutoDiff): # pragma: no cover
    
    # K3 = K³ + εI = MMt
    # K+ ≈ K K3^-1 K = (M^-1 K)t (M^-1 K)
    # K ≈ (K+ M) (K+ M)t
    
    # K.solve(b) = K3.quad(K, K b)
    # K.quad(b) = K3.quad(K b)
    # K.quad(b, c) = K3.quad(Kb, Kc)
    # K.logdet = 1/3 K3.logdet
    # K.correlate(b) = K3.quad(K, K K3.correlate(b))
    # K.correlate(b, T) = K3.correlate(K3.quad(K, K b), T)
    # K.decorrelate(b) = K3.decorrelate(K b)
    # K.decorrelate(b, T) = K @ K3.decorrelate(b, T)
    
    # TODO pass directly the instantiated decomposition of K^3 along with K?
    
    def __init__(self, K, decompcls, **kw):
        """
        
        Approximate a pseudoinverse using an inversion.
        
        Parameters
        ----------
        K :
            The matrix to be decomposed.
        decompcls : type
            Class used to decompose K^3, need not support ill-conditioned
            matrices.
        
        """
    
        self._K3 = decompcls(K @ K @ K, **kw)
        self._K = K
    
    def solve(self, b):
        K, K3 = self._K, self._K3
        return K3.quad(K, K @ b)
    
    def quad(self, b, c=None):
        K, K3 = self._K, self._K3
        if c is None:
            return K3.quad(K @ b)
        elif c.dtype == object:
            return K3.quad(K @ b, numpy.asarray(K) @ c)
        else:
            return K3.quad(K @ b, K @ c)
    
    def logdet(self):
        return self._K3.logdet() / 3
        # TODO this is not a pseudodeterminant
    
    def correlate(self, b, *, transpose=False):
        K, K3 = self._K, self._K3
        if transpose:
            return K3.correlate(K3.quad(K, K @ b), transpose=True)
        else:
            return K3.quad(K, K @ K3.correlate(b))
    
    def decorrelate(self, b, *, transpose=False):
        K, K3 = self._K, self._K3
        if transpose:
            return K @ K3.decorrelate(b, transpose=True)
        else:
            return K3.decorrelate(K @ b)
    
    @property
    def m(self):
        return self._K3.m
    
    @property
    def eps(self):
        return jnp.cbrt(self._K3.eps)

class Pinv2(DecompAutoDiff): # pragma: no cover
    
    # K3 = K³ + εI = MM'
    # K⁺ = ∑ n=1^∞ εⁿ⁻¹ K K3⁻ⁿ K =
    #    = ∑ n=1^∞ εⁿ⁻¹ (M⁻ⁿ K)' (M⁻ⁿ K)
    #    = A⁺' A⁺
    # A⁺ = [M⁻¹ K; √ε M⁻² K; ...]
    # A = (A⁺' A⁺)⁺ A⁺' = K A⁺' = K² [M'⁻¹, √ε M'⁻², ...]
    # K = AA'
        
    # TODO pass directly the instantiated decomposition of K^3 along with K?
    
    def __init__(self, K, decompcls, *, N=1, epsrel='auto', epsabs=0, **kw):
        """
        
        Approximate K⁺ with the truncated series
        
            K⁺ ≈ ∑ n=1^N εⁿ⁻¹ K (K³ + εI)⁻ⁿ K.
        
        Parameters
        ----------
        K : (n, n) array
            The matrix to be decomposed.
        decompcls : callable
            Class used to decompose K³ + εI, need not support ill-conditioned
            matrices. It should not add its own diagonal regularization.
        N : int
            Truncation order of the series.
        eps : float
            Positive regularization used to make K³ invertible, relative to
            an upper bound on the spectral radius of K³.
        **kw :
            Keyword arguments are passed to `decompcls`.
        
        """
        
        assert N >= 1
        K = jnp.asarray(K)
        K3 = K @ K @ K
        eps = self._parseeps(K3, epsrel, epsabs)
        K3 = K3.at[jnp.diag_indices_from(K3)].add(eps)
        self._K3 = decompcls(K3, **kw)
        self._K = K
        self._N = N
        self._eps = eps
    
    # K.solve(b) =
    # ∑ n=1^N ε^(n-1) K3.quad(K3.solve^n/2-1(K), K3.solve^n/2(Kb))       n even
    #                 K3.quad(K3.solve^(n-1)/2(K), K3.solve^(n-1)/2(Kb)) n odd
    # n  p1 p2
    # 1  0  0
    # 2  0  1
    # 3  1  1
    # 4  1  2
    # 5  2  2
    # etc.
    def solve(self, b):
        eps, K, K3, N = self._eps, self._K, self._K3, self._N
        K3solveK = K
        K3solveKb = K @ b
        coef = 1
        acc = K3.quad(K3solveK, K3solveKb)
        for n in range(2, N + 1):
            coef *= eps
            if n % 2:
                K3solveK = K3.solve(K3solveK)
            else:
                K3solveKb = K3.solve(K3solveKb)
            acc += coef * K3.quad(K3solveK, K3solveKb)
        return acc
    
    # K.quad(b) =
    # ∑ n=1^N ε^(n-1) K3.quad(K3.solve^n/2-1(Kb), K3.solve^n/2(Kb))   n even
    #                 K3.quad(K3.solve^(n-1)/2(Kb))                   n odd
    # K.quad(b, c) =
    # ∑ n=1^N ε^(n-1) K3.quad(K3.solve^n/2-1(Kb), K3.solve^n/2(Kc))       n even
    #                 K3.quad(K3.solve^(n-1)/2(Kb), K3.solve^(n-1)/2(Kc)) n odd
    # K.quad(b, c_obj) =
    # ∑ n=1^N ε^(n-1) K3.quad(K3.solve^(n-1)(Kb), Kc)
    #
    # n  p1 p2
    # 1  0  0
    # 2  0  1
    # 3  1  1
    # 4  1  2
    # 5  2  2
    # etc.
    def quad(self, b, c=None):
        eps, K, K3, N = self._eps, self._K, self._K3, self._N
        coef = 1
        
        if c is None:
            K3solveKb = K @ b
            acc = K3.quad(K3solveKb)
            for n in range(2, N + 1):
                coef *= eps
                if n % 2:
                    acc += coef * K3.quad(K3solveKb)
                else:
                    K3solveKb_next = K3.solve(K3solveKb)
                    acc += coef * K3.quad(K3solveKb, K3solveKb_next)
                    K3solveKb = K3solveKb_next

        elif c.dtype == object:
            K3solveKb = K @ b
            Kc = numpy.asarray(K) @ c
            acc = K3.quad(K3solveKb, Kc)
            eps = numpy.asarray(eps)
            for n in range(2, N + 1):
                coef *= eps
                K3solveKb = K3.solve(K3solveKb)
                acc += coef * K3.quad(K3solveKb, Kc)

        else:
            K3solveKc = K @ c
            K3solveKb = K @ b
            acc = K3.quad(K3solveKb, K3solveKc)
            for n in range(2, N + 1):
                coef *= eps
                if n % 2:
                    K3solveKb = K3.solve(K3solveKb)
                else:
                    K3solveKc = K3.solve(K3solveKc)
                acc += coef * K3.quad(K3solveKb, K3solveKc)

        return acc
    
    # K.logdet = 1/3 K3.logdet
    def logdet(self):
        return self._K3.logdet() / 3
        # TODO this is not a pseudodeterminant
    
    # K.correlate(b) =
    # K^2 ∑ n=1^N ε^(n-1)/2 K3.decorrelate^n(b_n, T)
    
    # K.correlate(b, T) = [
    #   ε^(n-1)/2 K3.decorrelate^n(K^2 b)
    #   for n in 1 ... N
    # ]
    def correlate(self, b, *, transpose=False):
        seps, K2, K3, N = jnp.sqrt(self._eps), self._K @ self._K, self._K3, self._N
        coef = 1
        
        if transpose:
            K3decK2b = K2 @ b
            parts = []
            for n in range(1, N + 1):
                K3decK2b = K3.decorrelate(K3decK2b)
                parts.append(coef * K3decK2b)
                coef *= seps
            return jnp.concatenate(parts, axis=max(0, b.ndim - 2))
        
        else:
            acc = 0
            for n in range(1, N + 1):
                K3decbn = b[K3.m * (n - 1):K3.m * n]
                for k in range(n):
                    K3decbn = K3.decorrelate(K3decbn, transpose=True)
                acc += coef * K3decbn
                coef *= seps
            return K2 @ acc
    
    # K.decorrelate(b) = [
    #   ε^(n-1)/2 K3.decorrelate^n(K b)
    #   for n in 1 ... N
    # ]
    
    # K.decorrelate(b, T) =
    # K ∑ n=1^N ε^(n-1)/2 K3.decorrelate^n(b_n, T)
    def decorrelate(self, b, *, transpose=False):
        seps, K, K3, N = jnp.sqrt(self._eps), self._K, self._K3, self._N
        coef = 1
        
        if transpose:
            acc = 0
            for n in range(1, N + 1):
                K3decbn = b[K3.m * (n - 1):K3.m * n]
                for k in range(n):
                    K3decbn = K3.decorrelate(K3decbn, transpose=True)
                acc += coef * K3decbn
                coef *= seps
            return K @ acc

        else:
            K3decKb = K @ b
            parts = []
            for n in range(1, N + 1):
                K3decKb = K3.decorrelate(K3decKb)
                parts.append(coef * K3decKb)
                coef *= seps
            return jnp.concatenate(parts, axis=max(0, b.ndim - 2))
    
    @property
    def m(self):
        return self._K3.m * self._N

    @property
    def eps(self):
        return jnp.cbrt(self._K3.eps)



# lsqfitgp/tests/test_pred.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import sys
import itertools
import time

import numpy as np
import gvar
from pytest import mark
import pytest

sys.path = ['.'] + sys.path
import lsqfitgp as lgp
import util

seed = np.random.SeedSequence(202302281256)
s, seed = seed.spawn(2)
rng = np.random.default_rng(s)

# TODO add gpkw for setting checksym, checkpos, solvers

def pred(seed, err, kw):
    rng = np.random.default_rng(seed)
    x = rng.uniform(-5, 5, size=20)
    xpred = rng.uniform(-10, 10, size=100)

    gp = lgp.GP(lgp.ExpQuad())
    gp.addx(x, 'data')
    gp.addx(xpred, 'pred')
    
    y = np.tanh(x)
    if err:
        datagp = lgp.GP(0.1 ** 2 * lgp.Cauchy(scale=0.3))
        datagp.addx(x, 'data')
        y = y + datagp.prior('data')
    
    result = gp.pred({'data': y}, 'pred', **kw)
    if isinstance(result, tuple) and len(result) == 2:
        mean, cov = result
    elif isinstance(result, np.ndarray):
        mean = gvar.mean(result)
        cov = gvar.evalcov(result)
    
    return mean, cov

@mark.parametrize('kw1,kw2', itertools.combinations([
    dict(fromdata=fromdata, raw=raw, keepcorr=keepcorr)
    for fromdata, raw, keepcorr in itertools.product([False, True], repeat=3)
    if not (raw and keepcorr)
], 2))
@mark.parametrize('err', [False, True])
def test_pred(err, kw1, kw2):
    if err and kw1['fromdata'] != kw2['fromdata']:
        pytest.skip()
    for _ in range(10):
        global seed
        s, seed = seed.spawn(2)
        m1, cov1 = pred(s, err, kw1)
        m2, cov2 = pred(s, err, kw2)
        util.assert_allclose(m1, m2, rtol=1e-5 if err else 1e-6)
        util.assert_close_matrices(cov1, cov2, rtol=1e-5 if err else 1e-1)
        # TODO 1e-1 looks too inaccurate

def test_double_pred():
    n = 50
    gp = lgp.GP(lgp.ExpQuad())
    ax, bx = rng.standard_normal((2, n))
    gp.addx(ax, 'a')
    gp.addx(bx, 'b')
    m = rng.standard_normal((n, n))
    ay = gvar.gvar(rng.standard_normal(n), m.T @ m)
    m1, cov1 = gp.predfromdata({'a': ay}, 'b', raw=True)
    m2, cov2 = gp.predfromfit(gp.predfromdata({'a': ay}, ['a']), 'b', raw=True)
    util.assert_allclose(m2, m1, rtol=1e-7)
    util.assert_close_matrices(cov2, cov1, rtol=1e-6)



import itertools

from matplotlib import pyplot as plt
import numpy as np
from scipy import special

def covfun(theta, tau, alpha, v):
    norm = special.beta(alpha, v + tau) / special.beta(alpha, v)
    return norm * special.hyp2f1(tau, alpha, alpha + v + tau, np.cos(theta))

fig, ax = plt.subplots(clear=True, num='hyp2f1')

taulist = [0.1, 1]
alphalist = [0.1, 1]
vlist = [0.5, 1]

theta = np.linspace(0, np.pi, 1000)

for tau, alpha, v in itertools.product(taulist, alphalist, vlist):
    y = covfun(theta, tau, alpha, v)
    ax.plot(theta, y, label=f'$\\tau={tau}, \\alpha={alpha}, \\nu={v}$')

ax.set_xlabel('$\\theta$ (geodesic distance)')
ax.set_ylabel('covariance')
ax.set_title('F-family covariance function on a sphere')
ax.legend()

fig.show()



import jax
from jax import core

t = []

def f(x):
    global t
    t.append(x)
    return x

g = jax.jit(f)
h = jax.grad(f)
i = jax.jit(jax.grad(f))
j = jax.vmap(f)
k = jax.grad(jax.grad(f))
l = jax.jacfwd(f)



# lsqfitgp/examples/r.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

                            EXAMPLE R.

    Where we decide to introduce a strong anisotropy despite
    evidence of its absence.

"""

import lsqfitgp as lgp
from lsqfitgp import _linalg
from matplotlib import pyplot as plt
import numpy as np
import gvar

xdata1d = np.linspace(-4, 4, 10)
xpred1d = np.linspace(-10, 10, 50)

def makegrid(array1d):
    x, y = np.meshgrid(array1d, array1d)
    out = np.empty(len(array1d) * len(array1d), [('x', float), ('y', float)])
    out['x'] = x.reshape(-1)
    out['y'] = y.reshape(-1)
    return out

xdata = makegrid(xdata1d)
xpred = makegrid(xpred1d)
z = np.cos(xdata['x']) * np.cos(xdata['y'])

gp = lgp.GP(lgp.ExpQuad(scale=3, dim='x') * lgp.ExpQuad(scale=1, dim='y'))
gp.addx(xdata, 'pere')
gp.addx(xpred, 'banane')

print('fit...')
m, cov = gp.predfromdata({'pere': z}, 'banane', raw=True)

print('samples...')
samples = m + gp.decompose(cov, solver='chol', epsrel=1e-5).correlate(np.random.randn(len(cov)))

print('plot...')
fig, ax = plt.subplots(num='r', clear=True, subplot_kw=dict(projection='3d', computed_zorder=False))

ax.scatter(xdata['x'], xdata['y'], z, color='black', zorder=10)
plotxpred = xpred.reshape(len(xpred1d), len(xpred1d))
ax.plot_surface(plotxpred['x'], plotxpred['y'], samples.reshape(plotxpred.shape), alpha=0.85, cmap='viridis')

ax.view_init(elev=60, azim=30)
for axis in 'xyz':
    exec(f'ax.set_{axis}label("{axis}")')

fig.show()



# lsqfitgp/examples/e.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

                            EXAMPLE E.

    Where observing both a function and its derivative put some
    restraint on their behaviour.

"""

import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np
import gvar

xdata = np.linspace(-5, 5, 10)
xpred = np.linspace(-15, 25, 200)
y = np.sin(xdata)
y[1::2] = np.cos(xdata[1::2])

print('make GP...')
gp = lgp.GP(lgp.ExpQuad(scale=3))
gp.addx(xdata[0::2], 'data', deriv=0)
gp.addx(xdata[1::2], 'dataderiv', deriv=1)
gp.addx(xpred, 'pred', deriv=0)
gp.addx(xpred, 'predderiv', deriv=1)

print('fit...')
u = gp.predfromdata({'data': y[0::2], 'dataderiv': y[1::2]}, ['pred', 'predderiv'])

print('figure...')
fig, ax = plt.subplots(num='e', clear=True)

colors = dict()
for label in u:
    m = gvar.mean(u[label])
    s = gvar.sdev(u[label])
    patch = ax.fill_between(xpred, m - s, m + s, label=label, alpha=0.5)
    colors[label] = patch.get_facecolor()[0]
    
print('samples...')
for sample in gvar.raniter(u, 30):
    for label in u:
        ax.plot(xpred, sample[label], '-', color=colors[label])

for deriv, marker in (0, '+'), (1, 'x'):
    ax.plot(xdata[deriv::2], y[deriv::2], f'k{marker}', label=f'data deriv {deriv}')
ax.legend(loc='best')

fig.show()



"""Test of double integral constraint"""

import lsqfitgp as lgp
import numpy as np
from matplotlib import pyplot as plt
import gvar

np.random.seed(20220417)

#### DEFINE MODEL ####
# h ~ GP
# f = h''
# int_0^1 dx f(x) = [h'(x)]_0^1 = h'(1) - h'(0)
# int_0^1 dx x f(x) = [xh'(x) - h(x)]_0^1 = h'(1) - h(1) + h(0)

gp = lgp.GP(lgp.ExpQuad())

x = np.linspace(0, 1, 10)
gp.addx(x, 'data', deriv=2)

gp.addx([0, 1], 'xinteg', deriv=1)
gp.addtransf({'xinteg': [-1, 1]}, 'integ')

gp.addx([0, 1], 'xintegx0')
gp.addx(1, 'xintegx1', deriv=1)
gp.addtransf({'xintegx1': 1, 'xintegx0': [1, -1]}, 'integx')

#### GENERATE FAKE DATA ####

prior = gp.predfromdata({
    'integ' : 1,
    'integx': 1,
}, ['data', 'integ', 'integx'])
priorsample = next(gvar.raniter(prior))

datamean = priorsample['data']
dataerr = np.full_like(datamean, 1)
datamean = datamean + dataerr * np.random.randn(*dataerr.shape)
data = gvar.gvar(datamean, dataerr)

# check the integral is one with trapezoid rule
print('prior:')
y = priorsample['data']
checksum = np.sum((      y[1:] +       y[:-1]) / 2 * np.diff(x))
print('sum_i int dx   f_i(x) =', checksum)
checksum = np.sum(((y * x)[1:] + (y * x)[:-1]) / 2 * np.diff(x))
print('sum_i int dx x f_i(x) =', checksum)

#### FIT ####

pred = gp.predfromdata({
    'integ' :    1,
    'integx':    1,
    'data'  : data,
}, ['data', 'integ', 'integx'])

# check the integral is one with trapezoid rule
print('posterior:')
y = pred['data']
checksum = np.sum((      y[1:] +       y[:-1]) / 2 * np.diff(x))
print('sum_i int dx   f_i(x) =', checksum)
checksum = np.sum(((y * x)[1:] + (y * x)[:-1]) / 2 * np.diff(x))
print('sum_i int dx x f_i(x) =', checksum)

#### PLOT RESULTS ####

fig, ax = plt.subplots(num='doubleint', clear=True)

y = pred['data']
m = gvar.mean(y)
s = gvar.sdev(y)
ax.fill_between(x, m - s, m + s, alpha=0.6)

y = priorsample['data']
ax.plot(x, y)

y = pred['data'] * x
m = gvar.mean(y)
s = gvar.sdev(y)
ax.fill_between(x, m - s, m + s, alpha=0.6)

y = priorsample['data'] * x
ax.plot(x, y)

ax.errorbar(x, datamean, dataerr, color='black', linestyle='', capsize=2)

fig.show()



# lsqfitgp/_special/_bernoulli.py
#
# Copyright (c) 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools

import numpy
from scipy import special
import jax
from jax import numpy as jnp
from jax.scipy import special as jspecial

def periodic_bernoulli(n, x):
    # TODO to make this jittable, hardcode size to 60 and truncate by writing
    # zeros
    n = int(n)
    bernoulli = special.bernoulli(n)
    k = numpy.arange(n + 1)
    binom = special.binom(n, k)
    coeffs = binom[::-1] * bernoulli
    x = x % 1
    cond = x < 0.5
    x = jnp.where(cond, x, 1 - x)
    out = jnp.polyval(coeffs, x)
    if n % 2 == 1:
        out = out * jnp.where(cond, 1, -1)
    return out

@functools.partial(jax.custom_jvp, nondiff_argnums=(0,))
def scaled_periodic_bernoulli(n, x):
    """ periodic Bernoulli polynomial scaled such that B_n(0) = ζ(n) """
    tau = 2 * jnp.pi
    lognorm = n * jnp.log(tau) - jspecial.gammaln(n + 1)
    norm = jnp.exp(lognorm) / 2
    cond = n < 60
    smalls = norm * periodic_bernoulli(n if cond else 1, x)
    # n -> ∞: -Re e^2πix / i^n
    # don't use 1j ** n because it is very inaccurate
    arg = tau * x
    sign = jnp.where((n // 2) % 2, 1, -1)
    larges = sign * jnp.where(n % 2, jnp.sin(arg), jnp.cos(arg))
    return jnp.where(cond, smalls, larges)

@scaled_periodic_bernoulli.defjvp
def _scaled_periodic_bernoulli_jvp(n, primals, tangents):
    x, = primals
    xt, = tangents
    primal = scaled_periodic_bernoulli(n, x)
    tangent = 2 * jnp.pi * scaled_periodic_bernoulli(n - 1, x) * xt
    return primal, tangent



"""Fit of parton distributions functions (PDFs)

Like pdf9, but without using Woodbury"""

import warnings

import lsqfitgp as lgp
import numpy as np
from jax import numpy as jnp
from scipy import linalg, interpolate
from matplotlib import pyplot as plt, gridspec
import gvar

np.random.seed(20230212)
# do not use np.random.default_rng because gvar uses the old numpy gen
warnings.filterwarnings('ignore', r'total derivative orders \(\d+, \d+\) greater than kernel minimum \(\d+, \d+\)')

#### DEFINITIONS ####

ndata = 3000 # number of datapoints

evnames = ['Sigma', 'V', 'V3', 'V8', 'V15', 'T3', 'T8', 'T15']
pnames  = evnames + ['g']
tpnames = ['xSigma'] + evnames[1:] + ['xg']

nflav = len(pnames)

# grid used for DGLAP evolution
grid = np.array([
    1.9999999999999954e-07, # start logspace
    3.034304765867952e-07,
    4.6035014748963906e-07,
    6.984208530700364e-07,
    1.0596094959101024e-06,
    1.607585498470808e-06,
    2.438943292891682e-06,
    3.7002272069854957e-06,
    5.613757716930151e-06,
    8.516806677573355e-06,
    1.292101569074731e-05,
    1.9602505002391748e-05,
    2.97384953722449e-05,
    4.511438394964044e-05,
    6.843744918967896e-05,
    0.00010381172986576898,
    0.00015745605600841445,
    0.00023878782918561914,
    0.00036205449638139736,
    0.0005487795323670796,
    0.0008314068836488144,
    0.0012586797144272762,
    0.0019034634022867384,
    0.0028738675812817515,
    0.004328500638820811,
    0.006496206194633799,
    0.009699159574043398,
    0.014375068581090129,
    0.02108918668378717,
    0.030521584007828916,
    0.04341491741702269,
    0.060480028754447364,
    0.08228122126204893,
    0.10914375746330703, # end logspace, start linspace
    0.14112080644440345,
    0.17802566042569432,
    0.2195041265003886,
    0.2651137041582823,
    0.31438740076927585,
    0.3668753186482242,
    0.4221667753589648,
    0.4798989029610255,
    0.5397572337880445,
    0.601472197967335,
    0.6648139482473823,
    0.7295868442414312,
    0.7956242522922756,
    0.8627839323906108,
    0.9309440808717544,
    1, # end linspace
])

# grid used for data
datagrid = grid[15:-1] # exclude 1 since f(1) = 0 and zero errors upset the fit
nx = len(datagrid)

# grid used for plot
gridinterp = interpolate.interp1d(np.linspace(0, 1, len(grid)), grid)
plotgrid = gridinterp(np.linspace(0, 1, 200))

# generated linear map PDF(grid) -> data
i = np.arange(ndata)[:, None, None]
j = np.arange(nx)[None, None, :]
intensity_diagonal = np.exp(-1/2 * (i / ndata - j / nx) ** 2 * ndata * nx)
intensity_flat = 1
intensity = 0.9 * intensity_diagonal + 0.1 * intensity_flat
intensity = np.broadcast_to(intensity, (ndata, nflav, nx))
dof = 3
M = intensity * np.random.chisquare(dof, intensity.shape) / dof

#### GAUSSIAN PROCESS ####
# Ti ~ GP   (i = 3, 8, 15)
#
# fi ~ GP with sdev ~ x to compensate the scale ~ 1/x   (i = 3, 8, 15)
# Vi = fi'
# f(1) - f(0) = 3
# f3(1) - f3(0) = 1
# f8(1) - f8(0) = 3
# f15(1) - f15(0) = 3
#
# f1 ~ GP   (without scale compensation)
# tf1(x) = x^(a+1)/(a+2) f1(x)   <--- scale comp. is x^(a+1) instead of x^a,
#                                     to avoid doing x^a with a < 0 in x = 0
# Sigma(x) = tf1'(x) / x    (such that x Sigma(x) ~ x^a)
# the same with f2, tf2, g
# tf12 = tf1 + tf2
# tf12(0) - tf12(1) = 1
#
# [Sigma, g, V*, T*](1) = 0

# transformation from evolution to flavor basis
hyperprior = {
    # correlation length of the priors at x = 1
    'log(scale_T3)':    np.log(gvar.gvar(0.5, 0.2)),
    'log(scale_T8)':    np.log(gvar.gvar(0.5, 0.2)),
    'log(scale_T15)':   np.log(gvar.gvar(0.5, 0.2)),
    'log(scale_V)':     np.log(gvar.gvar(0.5, 0.2)),
    'log(scale_V3)':    np.log(gvar.gvar(0.5, 0.2)),
    'log(scale_V8)':    np.log(gvar.gvar(0.5, 0.2)),
    'log(scale_V15)':   np.log(gvar.gvar(0.5, 0.2)),
    'log(scale_Sigma)': np.log(gvar.gvar(0.5, 0.2)),
    'log(scale_g)':     np.log(gvar.gvar(0.5, 0.2)),
    # prior variances of primitives
    'log(sigma_T3)':    np.log(gvar.gvar(0.5, 0.2)),
    'log(sigma_T8)':    np.log(gvar.gvar(0.5, 0.2)),
    'log(sigma_T15)':   np.log(gvar.gvar(0.5, 0.2)),
    'log(sigma_V)':     np.log(gvar.gvar(0.5, 0.2)),
    'log(sigma_V3)':    np.log(gvar.gvar(0.5, 0.2)),
    'log(sigma_V8)':    np.log(gvar.gvar(0.5, 0.2)),
    'log(sigma_V15)':   np.log(gvar.gvar(0.5, 0.2)),
    'log(sigma_Sigma)': np.log(gvar.gvar(0.5, 0.2)),
    'log(sigma_g)':     np.log(gvar.gvar(0.5, 0.2)),
    # exponents of x Sigma(x) and x g(x) for x -> 0
    'U(alpha_Sigma)': gvar.BufferDict.uniform('U', -0.5, 0.5),
    'U(alpha_g)'    : gvar.BufferDict.uniform('U', -0.5, 0.5),
}

def makegp(hp):
    gp = lgp.GP(checkpos=False, checksym=False, solver='chol', epsabs=1e-10, epsrel=0)
    
    def kernel(scale, *, prim=False):
        eps = grid[0]
        scalefun = lambda x: scale * (x + eps) # = 1 / log'(x)
        kernel = lgp.Gibbs(scalefun=scalefun)
        if prim:
            kernel = kernel.rescale(scalefun, scalefun)
        return kernel
    
    # define Ts and Vs
    for suffix in ['', '3', '8', '15']:
        if suffix != '':
            var = hp['sigma_T' + suffix] ** 2
            gp.addproc(var * kernel(hp['scale_T' + suffix]), 'T' + suffix)
        var = hp['sigma_V' + suffix] ** 2
        gp.addproc(var * kernel(hp['scale_V' + suffix], prim=True), 'f' + suffix)
        gp.addprocderiv(1, 'V' + suffix, 'f' + suffix)
    
    # define xSigma
    var = hp['sigma_Sigma'] ** 2
    gp.addproc(var * kernel(hp['scale_Sigma']), 'f1')
    a = hp['alpha_Sigma']
    gp.addprocrescale(lambda x: x ** (a + 1) / (a + 2), 'tf1', 'f1')
    gp.addprocderiv(1, 'xSigma', 'tf1')
    
    # define xg
    var = hp['sigma_g'] ** 2
    gp.addproc(var * kernel(hp['scale_g']), 'f2')
    b = hp['alpha_g']
    gp.addprocrescale(lambda x: x ** (b + 1) / (b + 2), 'tf2', 'f2')
    gp.addprocderiv(1, 'xg', 'tf2')
    
    # define primitive of xSigma + xg
    gp.addproctransf({'tf1': 1, 'tf2': 1}, 'tf12')
    
    # definite integrals
    for proc in ['tf12', 'f', 'f3', 'f8', 'f15']:
        gp.addx([0, 1], proc + '-endpoints', proc=proc)
        gp.addlintransf(lambda x: x[1] - x[0], [proc + '-endpoints'], proc + '-diff')
    
    # right endpoint
    for proc in tpnames:
        gp.addx(1, f'{proc}(1)', proc=proc)
    
    # define a matrix of PDF values over the x grid
    for proc in tpnames:
        gp.addx(datagrid, proc + '-datagrid', proc=proc)
    gp.addlintransf(lambda *args: jnp.stack(args), [proc + '-datagrid' for proc in tpnames], 'datagrid')

    # linear data
    gp.addtransf({'datagrid': M}, 'datalatent', axes=2)

    # define a matrix of PDF values over the plot grid
    for proc in tpnames:
        gp.addx(plotgrid, proc + '-plotgrid', proc=proc)
    gp.addlintransf(lambda *args: jnp.stack(args), [proc + '-plotgrid' for proc in tpnames], 'plotgrid')

    return gp

constraints = {
    'tf12-diff': 1,
    'f-diff'   : 3,
    'f3-diff'  : 1,
    'f8-diff'  : 3,
    'f15-diff' : 3,
    'xSigma(1)': 0,
    'V(1)'     : 0,
    'V3(1)'    : 0,
    'V8(1)'    : 0,
    'V15(1)'   : 0,
    'T3(1)'    : 0,
    'T8(1)'    : 0,
    'T15(1)'   : 0,
    'xg(1)'    : 0,
}

#### FAKE DATA ####

truehp = gvar.sample(hyperprior)

# rescale M to avoid having data depend almost uniquely on divergent functions
M[:, 0, :] /= datagrid ** truehp['alpha_Sigma']
M[:, -1, :] /= datagrid ** truehp['alpha_g']

truegp = makegp(truehp)
trueprior, trueprior_cov = truegp.predfromdata(constraints, ['datalatent', 'plotgrid'], raw=True)
# no gvars because it's slow with >1000 datapoints
truedata = lgp.sample(trueprior, trueprior_cov, eps=1e-10)

v = truedata['datalatent']
dataerr = np.full_like(v, 0.1 * (np.max(v) - np.min(v)))
data = gvar.make_fake_data(gvar.gvar(v, dataerr))
dataerrcov = gvar.evalcov(data)
datamean = gvar.mean(data)

def check_constraints(y):
    # integrate approximately with trapezoid rule
    integ = np.sum((y[:, 1:] + y[:, :-1]) / 2 * np.diff(plotgrid), 1)
    print(f'int dx x (Sigma(x) + g(x)) = {integ[0] + integ[-1]:.2g}')
    for i in range(1, 5):
        print(f'int dx {tpnames[i]}(x) = {integ[i]:.2g}')
    for i, name in enumerate(tpnames):
        print(f'{name}(1) = {y[i, -1]:.2g}')
    print()

print('\ncheck constraints in fake data:')
check_constraints(truedata['plotgrid'])

#### FIT ####

information = gvar.gvar(dict(datalatent=data, **constraints))

fitkw = dict(
    raises=False,
    jit=True,
    method='gradient',
    verbosity=3,
    # covariance='fisher',
    initial=truehp,
    # fix={'alpha_Sigma': True, 'alpha_g': True},
)
fit = lgp.empbayes_fit(hyperprior, makegp, information, **fitkw)

gp = makegp(gvar.mean(fit.p))
pred, predcov = gp.predfromdata(information, ['datalatent', 'plotgrid'], raw=True)
# use raw because with gvars it becomes slow above ~1000 datapoints

print('\ncheck constraints in fit:')
check_constraints(pred['plotgrid'])

#### PLOT RESULTS ####

legkw = dict(loc='best', title_fontsize='large')
figkw = dict(figsize=[11, 8], clear=True)#, layout='constrained')

plt.close('all')
figa = plt.figure(num='pdf10-a', **figkw)
figb = plt.figure(num='pdf10-b', **figkw)
figc = plt.figure(num='pdf10-c', **figkw)
dataax = figa.add_subplot(211)
axs = [figa.add_subplot(212)]
axs += [
    figb.add_subplot(211, sharex=axs[0]),
    figb.add_subplot(212, sharex=axs[0]),
]
hypax = figc.add_subplot()

for i in range(nflav):
    
    label = tpnames[i]
    if label in ['xSigma', 'xg', 'V']:
        ax = axs[0]
    elif label.startswith('T'):
        ax = axs[1]
    else:
        ax = axs[2]
    
    if label.startswith('x'):
        expon = fit.p['alpha_' + label[1:]]
        label += f' $\\sim x^{{{expon}}}$'
    
    ypdf = pred['plotgrid'][i, :]
    ypdfcov = predcov['plotgrid', 'plotgrid'][i, :, i, :]
    m = ypdf
    s = np.sqrt(np.diag(ypdfcov))
    ax.fill_between(plotgrid, m - s, m + s, label=label, alpha=0.4, facecolor=f'C{i}')
    
    ax.plot(plotgrid, truedata['plotgrid'][i], color=f'C{i}')
    
    ax.set_xscale('log')

for ax in axs:
    ax.axvline(datagrid[0], linestyle='--', color='black')

if np.max(np.abs(axs[0].get_ylim())) > 10:
    axs[0].set_yscale('symlog', linthresh=10, subs=[2, 3, 4, 5, 6, 7, 8, 9])

for ax in axs:
    ax.set_xlabel('x')
    ax.set_ylabel('PDF(x)')
    
for ax in axs:
    ax.legend(title='PDFs', **legkw)

ax = dataax
ax.set(xlabel='Datapoint index', ylabel='Datapoint value')

zero = truedata['datalatent']
x = np.arange(len(zero))

# decimate the data to be displayed
sl = np.s_[::len(x) // 250 + 1]
zero = zero[sl]
x = x[sl]

ax.plot(x, truedata['datalatent'][sl] - zero, drawstyle='steps-mid', color='black', label='truth')

d = datamean[sl] - zero
ax.errorbar(x, d, dataerr[sl], color='black', linestyle='', linewidth=1, capsize=2, label='data')

d = pred['datalatent'][sl] - zero
dcov = predcov['datalatent', 'datalatent'][sl, sl]
m = d
s = np.sqrt(np.diag(dcov))
ax.fill_between(x, m - s, m + s, step='mid', color='gray', alpha=0.8, label='fit', zorder=10)

ax.legend(title='Data', **legkw)

ax = hypax
ax.set(ylabel='Hyperparameter name', xlabel='Transformed hyperparameter value')

x = list(range(len(hyperprior)))
keys = list(hyperprior.keys())
yprior = list(hyperprior.values())
ypost = list(fit.p.values())
ytrue = list(truehp.values())

ax.set_yticks(x)
ax.set_yticklabels(keys)

m = gvar.mean(yprior)
s = gvar.sdev(yprior)
ax.fill_betweenx(x, m - s, m + s, label='prior', color='lightgray')
ax.errorbar(gvar.mean(ypost), x, xerr=gvar.sdev(ypost), label='posterior', color='black', linestyle='', capsize=3, marker='.')
ax.plot(ytrue, x, 'x', label='true value', color='red')

ax.legend(title='Hyperparameters', **legkw)

figa.show()
figb.show()
figc.show()



# lsqfitgp/examples/i.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

                            EXAMPLE I.

    Where, due to obscure political reasons, we insist on forgetting
    important details that a reasonable man's mind would be fond of
    recalling later.

"""

import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np
import gvar

xdata = np.linspace(0, 10, 10)
xpred = np.linspace(-15, 25, 300)
y = np.sin(xdata)
yerr = 0.1

gp = lgp.GP(lgp.ExpQuad(scale=3))
gp.addx(xdata, 'pere')
gp.addx(xpred, 'banane')

uy = gvar.gvar(y + yerr * np.random.randn(len(y)), yerr * np.ones_like(y))
u = gp.predfromdata({'pere': uy}, 'banane', keepcorr=False)
assert gvar.cov(u[0], uy[0]) == 0
m = gvar.mean(u)
s = gvar.sdev(u)
cov = gvar.evalcov(u)

fig, ax = plt.subplots(num='i', clear=True)

patch = ax.fill_between(xpred, m - s, m + s, label='pred', alpha=0.5)
color = patch.get_facecolor()[0]
simulated_lines = np.random.multivariate_normal(m, cov, size=10)
ax.plot(xpred, simulated_lines.T, '-', color=color)
ax.errorbar(xdata, gvar.mean(uy), yerr=gvar.sdev(uy), fmt='k.', label='data')
ax.legend(loc='best')

fig.show()



from scipy import optimize
import numpy as np

sk, yk = np.random.randn(2, 10, 1000)
invh = optimize.LbfgsInvHessProduct(sk, yk)

def todense(self):
    """ Rewrite LbfgsInvHessProduct.todense avoiding matrix @ matrix """
    s, y, n_corrs, rho = self.sk, self.yk, self.n_corrs, self.rho
    Hk = np.eye(*self.shape, dtype=self.dtype)

    for i in range(n_corrs):
        # H  <--  (I - rho sy') H (I - rho ys') + rho ss' =
        #       = H - rho Hys' - rho sy'H + rho^2 sy'Hys' + rho ss' =
        #       = H - rho Hys' - rho sy'H + rho(1 + rho y'Hy) ss'
        rhoHys = -rho[i] * np.outer(Hk @ y[i], s[i])
        Hk += (rho[i] * (1 + rho[i] * (y[i] @ Hk @ y[i]))) * np.outer(s[i], s[i])
        Hk += rhoHys
        Hk += rhoHys.T

    return Hk

def _matmat(self, X):
    """ Rewrite LbfgsInvHessProduct._matvec broadcasting on columns of X """
    s, y, n_corrs, rho = self.sk, self.yk, self.n_corrs, self.rho
    Q = np.array(X, dtype=self.dtype, copy=True)
    assert Q.ndim == 2

    alpha = np.empty((n_corrs, Q.shape[1]))

    for i in range(n_corrs-1, -1, -1):
        alpha[i] = rho[i] * np.dot(s[i], Q)
        Q = Q - alpha[i]*y[i][:, np.newaxis]

    R = Q
    for i in range(n_corrs):
        beta = rho[i] * np.dot(y[i], R)
        R = R + s[i][:, np.newaxis] * (alpha[i] - beta)

    return R

inv1 = invh.todense()
inv2 = todense(invh)
inv3 = _matmat(invh, np.eye(*invh.shape))
np.testing.assert_allclose(inv1, inv2, atol=0, rtol=1e-5)
np.testing.assert_allclose(inv1, inv3, atol=0, rtol=1e-5)

%timeit -r1 invh.todense()
%timeit -r1 todense(invh)
%timeit -r1 _matmat(invh, np.eye(*invh.shape))
%timeit -r1 invh.matmat(np.eye(*invh.shape))



# lsqfitgp/tests/test_array.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import sys
import itertools

import numpy as np
from numpy.lib import recfunctions
from jax import tree_util
from jax import numpy as jnp
import pytest
import pandas as pd
import polars as pl

sys.path.insert(0, '.')
import lsqfitgp as lgp
from lsqfitgp import _array
import util

rng = np.random.default_rng(202301231524)

def test_ellipsis():
    x = np.empty((2, 3), dtype=[('a', float, 4)])
    y = lgp.StructuredArray(x)
    z = y[..., 0]
    assert z.shape == y.shape[:1]
    assert z['a'].shape == y.shape[:1] + x.dtype.fields['a'][0].shape

def fill_array_at_random(x, rng):
    if x.dtype.names:
        for name in x.dtype.names:
            fill_array_at_random(x[name], rng)
    elif x.dtype == bool:
        x[...] = rng.integers(0, 2, x.shape)
    elif np.issubdtype(x.dtype, np.integer):
        info = np.iinfo(x.dtype)
        x[...] = rng.integers(info.min, info.max, x.shape, dtype=x.dtype, endpoint=True)
    elif np.issubdtype(x.dtype, np.floating):
        x[...] = 100 * rng.standard_normal(x.shape)
    elif np.issubdtype(x.dtype, np.character):
        basetype = x.dtype.type
        size = x.dtype.itemsize // np.dtype((basetype, 1)).itemsize
        total = x.size * size
        data = rng.integers(1, 128, total, dtype='u1')
        data = data.view((np.bytes_, size))
        data = data.astype((basetype, size))
        x[...] = data.reshape(x.shape)
    elif x.dtype == object:
        for idx in np.ndindex(*x.shape):
            x[idx] = dict(idx=idx)
    else:
        raise NotImplementedError

def random_array(shape, dtype, rng=rng):
    x = np.empty(shape, dtype)
    fill_array_at_random(x, rng)
    return x

def crosscheck_operation(op, *arrays, **kw):
    """
    arrays = structured numpy arrays
    applies op both to arrays and to the conversion to StructuredArrays, then
    converts back to numpy arrays and check the result is the same. Op can be
    a multivalued function.
    """
    
    r1 = op(*arrays, **kw)
    r2 = op(*map(lgp.StructuredArray, map(np.copy, arrays)), **kw)
    
    if not isinstance(r1, (tuple, list)):
        r1 = (r1,)
        r2 = (r2,)
    
    assert len(r1) == len(r2)
    for res1, res2 in zip(r1, r2):
    
        if isinstance(res1, lgp.StructuredArray):
            res1 = np.asarray(res1)
        if isinstance(res2, lgp.StructuredArray):
            res2 = np.asarray(res2)

        util.assert_equal(res1, res2)

def test_broadcast_to():
    op = np.broadcast_to # uses __array_function__
    dtypes = [
        [('f0', float)],
        'f,f',
        'f,d,f',
        'f,O',
        'S25,U10,?',
        [('a', 'f,f', (3, 1)), ('b', [('c', 'd,d'), ('e', '?', 15)])],
    ]
    shapes = [
        (),
        (0,),
        (1,),
        (1, 0),
        (0, 1),
        (10,),
        (1, 2, 3),
        (2, 3, 4),
    ]
    transfs = [
        lambda s: s,
        lambda s: (1,) + s,
        lambda s: (1, 1) + s,
        lambda s: (8,) + s,
        lambda s: tuple(7 if i == 1 else i for i in s),
        lambda s: (4,) + tuple(3 if i == 1 else i for i in s),
    ]
    for dtype, shape, transf in itertools.product(dtypes, shapes, transfs):
        array = random_array(shape, dtype)
        crosscheck_operation(op, array, shape=transf(shape))

def test_broadcast_arrays():
    op = np.broadcast_arrays # uses __array_function__
    dtypes = [
        [('f0', float)],
        'f,f',
        'f,d,f',
        'f,O',
        'S25,U10,?',
        [('a', 'f,f', (3, 1)), ('b', [('c', 'd,d'), ('e', '?', 15)])],
    ]
    shapes = [
        (),
        (0,),
        (1,),
        (1, 0),
        (0, 1),
        (10,),
        (1, 2, 3),
        (2, 3, 4),
    ]
    transfs = [
        lambda s: s,
        lambda s: (1,) + s,
        lambda s: (1, 1) + s,
        lambda s: (8,) + s,
        lambda s: tuple(7 if i == 1 else i for i in s),
        lambda s: (4,) + tuple(3 if i == 1 else i for i in s),
    ]
    for dtype, shape, transf in itertools.product(dtypes, shapes, transfs):
        array1 = random_array(shape, dtype)
        array2 = random_array(transf(shape), dtype)
        crosscheck_operation(op, array1, array2)

def test_broadcast():
    def op(array1, array2):
        return lgp.broadcast(array1, array2).shape
    dtypes = [
        [('f0', float)],
        'f,f',
        'f,d,f',
        'f,O',
        'S25,U10,?',
        [('a', 'f,f', (3, 1)), ('b', [('c', 'd,d'), ('e', '?', 15)])],
    ]
    shapes = [
        (),
        (0,),
        (1,),
        (1, 0),
        (0, 1),
        (10,),
        (1, 2, 3),
        (2, 3, 4),
    ]
    transfs = [
        lambda s: s,
        lambda s: (1,) + s,
        lambda s: (1, 1) + s,
        lambda s: (8,) + s,
        lambda s: tuple(7 if i == 1 else i for i in s),
        lambda s: (4,) + tuple(3 if i == 1 else i for i in s),
    ]
    for dtype, shape, transf in itertools.product(dtypes, shapes, transfs):
        array1 = random_array(shape, dtype)
        array2 = random_array(transf(shape), dtype)
        crosscheck_operation(op, array1, array2)

def test_asarray():
    def op(array):
        return lgp.asarray(array)
    dtypes = [
        [('f0', float)],
        'f,f',
        'f,d,f',
        'f,O',
        'S25,U10,?',
        [('a', 'f,f', (3, 1)), ('b', [('c', 'd,d'), ('e', '?', 15)])],
    ]
    shapes = [
        (),
        (0,),
        (1,),
        (1, 0),
        (0, 1),
        (10,),
        (1, 2, 3),
        (2, 3, 4),
    ]
    for dtype, shape in itertools.product(dtypes, shapes):
        array = random_array(shape, dtype)
        crosscheck_operation(op, array)

def test_multiindex():
    def op(a):
        names = list(a.dtype.names)
        return a[names], a[names[:2]], a[names[-2:]], a[names[:1]]
    dtypes = [
        [('f0', float)],
        'f,f',
        'f,d,f',
        'f,O',
        'S25,U10,?',
        [('a', 'f,f', (3, 1)), ('b', [('c', 'd,d'), ('e', '?', 15)])],
    ]
    shapes = [
        (),
        (0,),
        (1,),
        (1, 0),
        (0, 1),
        (10,),
        (1, 2, 3),
        (2, 3, 4),
    ]
    for dtype, shape in itertools.product(dtypes, shapes):
        array = random_array(shape, dtype)
        crosscheck_operation(op, array)

def array_meta_hash(a):
    d = a.dtype
    shape_hash = sum(a.shape) + a.size + a.ndim
    type_hash = len(d) + d.itemsize + sum(d[i].num for i in range(len(d))) + d.num
    return shape_hash + type_hash

def test_setfield():
    def op(a):
        name = a.dtype.names[0]
        a0 = a[name]
        rng = np.random.default_rng(array_meta_hash(a))
        val = random_array(a0.shape, a0.dtype, rng)
        if hasattr(a, 'at'):
            a = a.at[name].set(val)
        else:
            a[name] = val
        return a
    dtypes = [
        [('f0', float)],
        'f,f',
        'f,d,f',
        'f,O',
        'S25,U10,?',
        [('a', 'f,f', (3, 1)), ('b', [('c', 'd,d'), ('e', '?', 15)])],
    ]
    shapes = [
        (),
        (0,),
        (1,),
        (1, 0),
        (0, 1),
        (10,),
        (1, 2, 3),
        (2, 3, 4),
    ]
    for dtype, shape in itertools.product(dtypes, shapes):
        array = random_array(shape, dtype)
        crosscheck_operation(op, array)

def test_tree():
    dtypes = [
        [('f0', float)],
        'f,f',
        'f,d,f',
        'f,O',
        'S25,U10,?',
        [('a', 'f,f', (3, 1)), ('b', [('c', 'd,d'), ('e', '?', 15)])],
    ]
    shapes = [
        (),
        (0,),
        (1,),
        (1, 0),
        (0, 1),
        (10,),
        (1, 2, 3),
        (2, 3, 4),
    ]
    for dtype, shape in itertools.product(dtypes, shapes):
        array1 = random_array(shape, dtype)
        array = lgp.StructuredArray(array1)
        children, aux_data = tree_util.tree_flatten(array)
        array = tree_util.tree_unflatten(aux_data, children)
        array2 = np.asarray(array)
        util.assert_equal(array1, array2)

def test_tree_reshaped():
    dtypes = [
        [('f0', float)],
        'f,5f',
        'f,d,f',
        'f,O',
        'S25,U10,?',
        [('a', 'f,f', (3, 1)), ('b', [('c', 'd,d'), ('e', '?', 15)])],
    ]
    shapes = [
        (),
        (0,),
        (1,),
        (1, 0),
        (0, 1),
        (10,),
        (1, 2, 3),
        (2, 3, 4),
    ]
    for dtype, shape in itertools.product(dtypes, shapes):
        
        basearray = random_array(shape, dtype)
        
        slices = tuple(slice(rng.integers(size + 1)) for size in shape)
        array1 = basearray[slices]
        
        array = lgp.StructuredArray(basearray)
        children, aux_data = tree_util.tree_flatten(array)
        children = tuple(x[slices] for x in children)
        array = tree_util.tree_unflatten(aux_data, children)
        array2 = np.asarray(array)
        
        util.assert_equal(array1, array2)

        lead = (3,)
        array1 = np.broadcast_to(basearray, lead + shape)
        
        array = lgp.StructuredArray(basearray)
        children, aux_data = tree_util.tree_flatten(array)
        children = tuple(np.broadcast_to(x, lead + x.shape) for x in children)
        array = tree_util.tree_unflatten(aux_data, children)
        array2 = np.asarray(array)
        
        util.assert_equal(array1, array2)

        if shape and shape[0]:

            array1 = basearray[0]
            
            array = lgp.StructuredArray(basearray)
            children, aux_data = tree_util.tree_flatten(array)
            children = tuple(x[0] for x in children)
            array = tree_util.tree_unflatten(aux_data, children)
            array2 = np.asarray(array)
            
            util.assert_equal(array1, array2)

def test_reshape():
    def op1(array, shape):
        return array.reshape(shape)
    def op2(array, shape):
        return array.reshape(*shape) if shape else array.reshape(shape)
    dtypes = [
        [('f0', float)],
        'f,f',
        'f,d,f',
        'f,O',
        'S25,U10,?',
        [('a', 'f,f', (3, 1)), ('b', [('c', 'd,d'), ('e', '?', 15)])],
    ]
    shapes = [
        (),
        (0,),
        (1,),
        (1, 0),
        (0, 1),
        (10,),
        (1, 2, 3),
        (2, 3, 4),
    ]
    transfs = [
        lambda s: s,
        lambda s: (1,) + s,
        lambda s: s + (1,),
        lambda s: (1, 1) + s,
        lambda s: (-1,) + s,
        lambda s: s + (-1,),
        lambda s: s[:-1] + (-1,),
        lambda s: s[:-2] + (-1,),
        lambda s: (-1,) + s[1:],
        lambda s: (-1,) + s[2:],
    ]
    for dtype, shape, transf in itertools.product(dtypes, shapes, transfs):
        array = random_array(shape, dtype)
        newshape = transf(shape)
        if -1 in newshape and (not shape or 0 in shape):
            continue
        crosscheck_operation(op1, array, shape=newshape)
        crosscheck_operation(op2, array, shape=newshape)

def test_repr():
    x = np.array(
      [[(  79.4607   ,   34.16494104,  True),
        (-122.71211  ,   70.48242559, False)],
       [( -95.21081  ,   80.3448    , False),
        (  52.144142 ,   57.35083659, False)],
       [(  97.3018   ,  -65.78833393,  True),
        (  88.46741  ,  -85.54110175, False)],
       [(  -2.7272243,  -16.86393124, False),
        (   0.946183 ,  -50.12316733, False)],
       [( -19.002777 , -114.6684194 ,  True),
        ( -59.26707  ,   87.90692279, False)]],
      dtype=[('f0', '<f4'), ('f1', '<f8'), ('f2', '?')]
    )
    y = lgp.StructuredArray(x)
    s = """\
StructuredArray({
    'f0':
    array([[  79.4607   , -122.71211  ],
           [ -95.21081  ,   52.144142 ],
           [  97.3018   ,   88.46741  ],
           [  -2.7272243,    0.946183 ],
           [ -19.002777 ,  -59.26707  ]], dtype=float32),
    'f1':
    array([[  34.16494104,   70.48242559],
           [  80.3448    ,   57.35083659],
           [ -65.78833393,  -85.54110175],
           [ -16.86393124,  -50.12316733],
           [-114.6684194 ,   87.90692279]]),
    'f2':
    array([[ True, False],
           [False, False],
           [ True, False],
           [False, False],
           [ True, False]]),
})"""
    assert repr(y) == s
    
    x = np.array(
      [(43.52967   , -258.21209855, False),
       ( 0.80091816,   80.99479392, False)],
      dtype=[('f0', '<f4'), ('f1', '<f8'), ('f2', '?')]
    )
    y = lgp.StructuredArray(x)
    s = "StructuredArray({'f0': array([43.52967   ,  0.80091816], dtype=float32), 'f1': array([-258.21209855,   80.99479392]), 'f2': array([False, False])})"
    assert repr(y) == s

    x = np.array(
      [([ 102.43267  ,   25.42825  ,  -76.084114 ,   39.536003 ],   85.10061608),
       ([  -6.0310173, -108.25371  ,  -20.205214 , -233.0124   ], -152.64632798)],
      dtype=[('f0', '<f4', (4,)), ('f1', '<f8')]
    )
    y = lgp.StructuredArray(x)
    s = """\
StructuredArray({
    'f0':
    array([[ 102.43267  ,   25.42825  ,  -76.084114 ,   39.536003 ],
           [  -6.0310173, -108.25371  ,  -20.205214 , -233.0124   ]],
          dtype=float32),
    'f1': array([  85.10061608, -152.64632798]),
})"""
    assert repr(y) == s
    
    x = np.zeros(5, [])
    y = lgp.StructuredArray(x)
    s = "StructuredArray({})"
    assert repr(y) == s

def test_set_subfield():
    y = random_array(5, [('a', [('b', float)])])
    x = lgp.StructuredArray(y)
    a = x['a']
    x = x.at['a'].set(random_array(a.shape, a.dtype))
    b = x['a']['b']
    newb = random_array(b.shape, b.dtype)
    x1 = x.at['a'].set(x['a'].at['b'].set(newb))
    x2 = x.at['a']['b'].set(newb)
    util.assert_equal(np.array(x1), np.array(x2))

def test_squeeze():
    dtypes = [
        [('f0', float)],
        'f,f',
        'f,d,f',
        # 'f,O',
        # structured_to_unstructured does not work on object arrays, see
        # numpy issue #21990
        'S25,U10,?',
        [('a', 'f,f', (3, 1)), ('b', [('c', 'd,d'), ('e', '?', 15)])],
        [('a', [('aa', 'f,f', 2), ('ab', float)], 3)],
    ]
    shapes = [
        (),
        (0,),
        (1,),
        (1, 0),
        (0, 1),
        (10, 1),
        (1, 2, 3),
        (2, 3, 4, 1),
    ]
    for dtype, shape in itertools.product(dtypes, shapes):
        array = random_array(shape, dtype)
        crosscheck_operation(np.squeeze, array)

    shapes = [
        (1,),
        (1, 0),
        (1, 2),
        (1, 3, 7),
    ]
    for dtype, shape in itertools.product(dtypes, shapes):
        array = random_array(shape, dtype)
        crosscheck_operation(lambda a: a.squeeze(axis=0), array)
        crosscheck_operation(lambda a: a.squeeze(axis=(0,)), array)
    
    shapes = [
        (1, 0, 1),
        (1, 2, 1),
        (1, 3, 1, 7),
    ]
    for dtype, shape in itertools.product(dtypes, shapes):
        array = random_array(shape, dtype)
        crosscheck_operation(lambda a: a.squeeze(axis=(0, 2)), array)
        crosscheck_operation(lambda a: a.squeeze(axis=(2, 0)), array)

@pytest.mark.parametrize('cls', [pl.DataFrame, pd.DataFrame])
def test_dataframe(cls):
    """ test DataFrame -> StructuredArray, both with Pandas and Polars """
    a = random_array(10, 'f4,f8,i1,i2,i4,i8,u1,u2,u4,u8,?,U16,S16')
    df = cls(pd.DataFrame.from_records(a))
    s = lgp.StructuredArray.from_dataframe(df)
    a2 = np.array(s)
    util.assert_equal(a, a2)

def test_not_handled():
    with pytest.raises(TypeError):
        np.concatenate([lgp.StructuredArray(np.zeros(1, 'd,d'))])

def test_asjax():
    
    x = random_array(4, float)
    assert isinstance(x, np.ndarray)
    x = _array._asarray_jaxifpossible(x)
    assert isinstance(x, jnp.ndarray)

    x = random_array(5, 'S5')
    x = _array._asarray_jaxifpossible(x)
    assert isinstance(x, np.ndarray)

    x = random_array(2, 'd,U10')
    x = _array._asarray_jaxifpossible(x)
    assert isinstance(x['f0'], jnp.ndarray)
    assert isinstance(x['f1'], np.ndarray)

def test_shortkey():
    x = random_array((2, 3), 'd,d')
    x = lgp.StructuredArray(x)
    elem = x[0]
    assert elem.shape == (3,)
    assert elem.dtype == 'd,d'

def test_longkey():
    x = random_array((2, 3), '5d,5d')
    x = lgp.StructuredArray(x)
    with pytest.raises(IndexError):
        elem = x[0, 1, 2]

def test_unflatten_dummy():
    x = random_array((2, 3), 'f,d,?')
    x = lgp.StructuredArray(x)
    _, aux = tree_util.tree_flatten(x)
    children = (8., False, None)
    y = tree_util.tree_unflatten(aux, children)
    assert y.shape == ()
    assert y.dtype == [('f0', float), ('f1', bool), ('f2', object)]

def test_incompatible_shapes():
    x = random_array((2, 3), 'f,2d,?')
    x = lgp.StructuredArray(x)
    _, aux = tree_util.tree_flatten(x)
    children = (8., False, None)
    y = tree_util.tree_unflatten(aux, children)
    assert y.shape == ()
    assert y.dtype == [('f0', float), ('f1', bool), ('f2', object)]

def test_ix():
    x = lgp.StructuredArray(random_array(8, 'f,2d,?'))
    y = lgp.StructuredArray(random_array(3, 'i,i'))
    z = random_array(7, 'd')

    x1, = np.ix_(x)
    assert isinstance(x1, lgp.StructuredArray)
    util.assert_equal(x1, x)

    x1, y1 = np.ix_(x, y)
    assert x1.shape == x.shape + (1,)
    assert y1.shape == (1,) + y.shape
    assert isinstance(x1, lgp.StructuredArray)
    assert isinstance(y1, lgp.StructuredArray)
    util.assert_equal(x1.squeeze(), x)
    util.assert_equal(y1.squeeze(), y)

    x1, y1, z1 = np.ix_(x, y, z)
    assert isinstance(x1, lgp.StructuredArray)
    assert isinstance(y1, lgp.StructuredArray)
    assert x1.shape == x.shape + (1, 1)
    assert y1.shape == (1,) + y.shape + (1,)
    assert z1.shape == (1, 1) + z.shape
    util.assert_equal(x1.squeeze(), x)
    util.assert_equal(y1.squeeze(), y)
    util.assert_equal(z1.squeeze(), z)

    z1, x1 = np.ix_(z, x)
    assert z1.shape == z.shape + (1,)
    assert x1.shape == (1,) + x.shape
    assert isinstance(x1, lgp.StructuredArray)
    util.assert_equal(z1.squeeze(), z)
    util.assert_equal(x1.squeeze(), x)

def test_structured_to_unstructured():
    dtypes = [
        [('a', float)],
        'f,f',
        'S25,U10,?',
        [('a', 'f,f')],
        [('a', float, 2)],
        [('a', float, (2, 3))],
        [('a', 'f,f', 3)],
        [('a', 'f,f', 3), ('b', 'f,f', 3)],
        [('a', 'f,f', (3, 5))],
        [('a', 'f,13f', (3, 5))],
        [('a', [('aa', 'f,2f', 3)], 5), ('b', float)],
        [('b', float), ('a', [('aa', [('aaa', 'f,2f', 3)], 5)], 7)],
        # structured_to_unstructured does not work on object arrays, see
        # numpy issue #21990
    ]
    shapes = [
        (),
        (0,),
        (1,),
        (1, 0),
        (0, 1),
        (7,),
        (7, 11),
    ]
    for dtype, shape in itertools.product(dtypes, shapes):
        array = random_array(shape, dtype)
        crosscheck_operation(recfunctions.structured_to_unstructured, array)

def test_unstructured_to_structured():
    dtypes = [
        [('a', float)],
        'f,f',
        'S25,U10',
        [('a', 'f,f')],
        [('a', float, 2)],
        [('a', float, (2, 3))],
        [('a', 'f,f', 3)],
        [('a', 'f,f', 3), ('b', 'f,f', 3)],
        [('a', 'f,f', (3, 5))],
        [('a', 'f,13f', (3, 5))],
        [('a', [('aa', 'f,2f', 3)], 5), ('b', float)],
        [('b', float), ('a', [('aa', [('aaa', 'f,2f', 3)], 5)], 7)],
        # structured_to_unstructured does not work on object arrays, see
        # numpy issue #21990
    ]
    shapes = [
        (),
        (0,),
        (1,),
        (1, 0),
        (0, 1),
        (7,),
        (7, 11),
    ]
    for dtype, shape in itertools.product(dtypes, shapes):
        x = random_array(shape, dtype)
        y = recfunctions.structured_to_unstructured(x)
        z = lgp.unstructured_to_structured(y, x.dtype)
        assert isinstance(z, lgp.StructuredArray)
        util.assert_equal(x, z)

def test_unstructured_to_structured_2():
    x = random_array((8, 9), float)
    y = lgp.unstructured_to_structured(x)
    assert isinstance(y, lgp.StructuredArray)
    assert y.shape == (8,)
    assert len(y.dtype) == 9
    z = recfunctions.structured_to_unstructured(y)
    util.assert_equal(x, z)
    x[0, 0] = 100000
    assert y['f0'][0] == 100000

    x = random_array((), float)
    with pytest.raises(ValueError):
        y = lgp.unstructured_to_structured(x)

def test_empty():
    like = lgp.StructuredArray(np.empty(0, 'f,f'))
    dtype = np.dtype([
        ('a', float),
        ('b', float, (2, 3)),
        ('c', 'f,d'),
        ('d', [('e', float, (4, 5))], (2, 3)),
        ('g', 'U8'),
    ])
    x = np.empty((4, 5), dtype, like=like)
    assert isinstance(x, lgp.StructuredArray)
    assert x.shape == (4, 5)
    assert x.dtype == dtype

    y = np.empty_like(x)
    assert isinstance(y, lgp.StructuredArray)
    assert y.shape == x.shape
    assert y.dtype == x.dtype

    shape = (8,)
    y = np.empty_like(x, shape=shape)
    assert isinstance(y, lgp.StructuredArray)
    assert y.shape == shape
    assert y.dtype == x.dtype

    dtype = np.dtype('f,2d')
    y = np.empty_like(x, dtype=dtype)
    assert isinstance(y, lgp.StructuredArray)
    assert y.shape == x.shape
    assert y.dtype == dtype

def test_nd():
    nd = lambda d: _array._nd(np.dtype(d))
    assert nd(float) == 1
    assert nd('f,f') == 1 + 1
    assert nd('f,2f') == 1 + 2
    assert nd([
        ('a', 'f,f', (3, 4)),
    ]) == (1 + 1) * 3 * 4
    assert nd([
        ('a', 'f,2f', (3, 4)),
    ]) == (1 + 2) * 3 * 4
